#!/usr/bin/env python2
#   IRPF90 is a Fortran90 preprocessor written in Python for programming using
#   the Implicit Reference to Parameters (IRP) method.
#   Copyright (C) 2009 Anthony SCEMAMA 
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License along
#   with this program; if not, write to the Free Software Foundation, Inc.,
#   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
#   Anthony Scemama
#   LCPQ - IRSAMC - CNRS
#   Universite Paul Sabatier
#   118, route de Narbonne      
#   31062 Toulouse Cedex 4      
#   scemama@irsamc.ups-tlse.fr

from command_line import command_line
import irpf90_t

def run():
  template = """
program codelet_%(name)s
  implicit none
  integer :: i
  double precision :: ticks_0, ticks_1, cpu_0, cpu_1
  integer, parameter :: irp_imax = %(NMAX)d

  %(precondition)s

  call provide_%(name)s

  double precision :: irp_rdtsc

  call cpu_time(cpu_0)
  ticks_0 = irp_rdtsc()
  do i=1,irp_imax
    call bld_%(name)s
  enddo
  ticks_1 = irp_rdtsc()
  call cpu_time(cpu_1)
  print *, '%(name)s'
  print *, '-----------'
  print *, 'Cycles:'
  print *,  (ticks_1-ticks_0)/dble(irp_imax)
  print *, 'Seconds:'
  print *,  (cpu_1-cpu_0)/dble(irp_imax)
end

  """

  name, NMAX, precondition, filename = command_line.codelet
  if precondition is None:
    precondition = ""
  else:
    precondition = "PROVIDE "+precondition
  file = open(filename,'w')
  file.write(template%locals())
  file.close()





#!/usr/bin/env python2
#   IRPF90 is a Fortran90 preprocessor written in Python for programming using
#   the Implicit Reference to Parameters (IRP) method.
#   Copyright (C) 2009 Anthony SCEMAMA 
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License along
#   with this program; if not, write to the Free Software Foundation, Inc.,
#   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
#   Anthony Scemama
#   LCPQ - IRSAMC - CNRS
#   Universite Paul Sabatier
#   118, route de Narbonne      
#   31062 Toulouse Cedex 4      
#   scemama@irsamc.ups-tlse.fr


import sys
from irpf90_t import *
from command_line import command_line
do_warnings = command_line.do_warnings

######################################################################
def fail(line,message):
  print """
Error:
-----
"""
  print message, '\n'
  if line is not None:
    assert isinstance(line,Line)
    print "file %s ; line %d :\n %s"%(line.filename,line.i,line.text)
  sys.exit(1)


######################################################################
def warn(line,message):
  if not command_line.do_warnings:
    return
  if line is not None:
    assert isinstance(line,Line)
    print """
Warning:
-------
"""
    print message, '\n'
    print "file %s, line %d:\n %s"%(line.filename,line.i,line.text)
  else:
    print "Warning: %s"%(message)


######################################################################
if __name__ == '__main__':
  line = Empty_line(3,"empty", "testfile")
  fail(line, "Message")




#!/usr/bin/env python2
#   IRPF90 is a Fortran90 preprocessor written in Python for programming using
#   the Implicit Reference to Parameters (IRP) method.
#   Copyright (C) 2009 Anthony SCEMAMA 
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License along
#   with this program; if not, write to the Free Software Foundation, Inc.,
#   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
#   Anthony Scemama
#   LCPQ - IRSAMC - CNRS
#   Universite Paul Sabatier
#   118, route de Narbonne      
#   31062 Toulouse Cedex 4      
#   scemama@irsamc.ups-tlse.fr


from irpf90_t import *
from util import *
import error

class Sub(object):

  ############################################################
  def __init__(self,text):
    assert type(text) == list
    assert len(text) > 0
    assert type(text[0]) in [Subroutine, Function]
    self.text = text
    self.called_by = []

  ############################################################
  def name(self):
    '''Name is lowercase'''
    if '_name' not in self.__dict__:
      self._name = find_subname(self.line)
    return self._name
  name = property(name)

  ############################################################
  def is_function(self):
    if '_is_function' not in self.__dict__:
      self._is_function = "function" in self.line.lower
    return self._is_function
  is_function = property(is_function)

  ############################################################
  def doc(self):
    if '_doc' not in self.__dict__:
      def f(l): return 
      buffer = filter(lambda l:type(l) == Doc, self.text)
      self._doc = map(lambda l: l.text.lstrip()[1:], buffer)
      if buffer == []:
        error.warn(None,"Subroutine %s is not documented"%(self.name))
    return self._doc
  doc = property(doc)

  ############################################################
  def line(self):
    if '_line' not in self.__dict__:
      self._line = self.text[0]
    return self._line
  line = property(line)

  ############################################################
  def touches(self):
    if '_touches' not in self.__dict__:
      from subroutines import subroutines
      self._touches = []
      for line in filter(lambda x: type(x) in [Touch, SoftTouch],self.text):
        self._touches += line.text.split()[1:]
      for sub in self.calls:
        if sub in subroutines:
          self._touches += subroutines[sub].touches
      self._touches = make_single(self._touches)
    return self._touches
  touches = property(touches)

  ############################################################
  def needs(self):
    if '_needs' not in self.__dict__:
      import parsed_text
    self._needs = make_single(self._needs)
    return self._needs
  needs = property(needs)

  ############################################################
  def to_provide(self):
    if '_to_provide' not in self.__dict__:
      import parsed_text
    return self._to_provide
  to_provide = property(to_provide)

  ############################################################
  def regexp(self):
    if '_regexp' not in self.__dict__:
      import re
      self._regexp = re.compile( \
        r"([^a-z0-9'\"_]|^)%s([^a-z0-9_]|$)"%(self.name),re.I)
    return self._regexp
  regexp = property(regexp)

  ############################################################
  def calls(self):
    if '_calls' not in self.__dict__:
      buffer = filter(lambda x: type(x) == Call,self.text)
      self._calls = []
      for line in buffer:
        sub = line.text.split('(',1)[0].split()[1].lower()
        self._calls.append(sub)
      self._calls = make_single(self._calls)
    return self._calls
  calls = property(calls)

######################################################################
if __name__ == '__main__':
  from preprocessed_text import preprocessed_text
  from variables import variables
  from subroutines import subroutines
  print map(lambda x: variables[x].needs, subroutines['full_ci'].needs)
  print subroutines['full_ci'].calls



#!/usr/bin/env python2
#   IRPF90 is a Fortran90 preprocessor written in Python for programming using
#   the Implicit Reference to Parameters (IRP) method.
#   Copyright (C) 2009 Anthony SCEMAMA 
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License along
#   with this program; if not, write to the Free Software Foundation, Inc.,
#   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
#   Anthony Scemama
#   LCPQ - IRSAMC - CNRS
#   Universite Paul Sabatier
#   118, route de Narbonne      
#   31062 Toulouse Cedex 4      
#   scemama@irsamc.ups-tlse.fr

import os
from zlib import crc32

irpdir = "IRPF90_temp/"
mandir = "IRPF90_man/"
irp_id = abs(crc32(os.getcwd()))


class Line(object):
  def __init__(self,i,text,filename):
    self.i = i
    self._text = None
    self.filename = filename
    self._lower = None
    self.set_text(text)

  def get_text(self):
    return self._text
  
  def set_text(self,value):
    self._text = value
    self._lower = value.lower()

  def get_lower(self):
    return self._lower

  text = property(fget=get_text, fset=set_text)
  lower = property(fget=get_lower)

class Empty_line(Line):
  str="Empty_line"
  def __init__(self,i,text,filename):
    Line.__init__(self,i,text,filename)
  def __repr__(self):
    return "%20s:%5d : %s"%("Empty_line",self.i,self.text)

class Simple_line(Line):
  str="Simple_line"
  def __init__(self,i,text,filename):
    Line.__init__(self,i,text,filename)
  def __repr__(self):
    return "%20s:%5d : %s"%("Simple_line",self.i,self.text)

class Declaration(Line):
  str="Declaration"
  def __init__(self,i,text,filename):
    Line.__init__(self,i,text,filename)
  def __repr__(self):
    return "%20s:%5d : %s"%("Declaration",self.i,self.text)

class Continue(Line):
  str="Continue"
  def __init__(self,i,text,filename):
    Line.__init__(self,i,text,filename)
  def __repr__(self):
    return "%20s:%5d : %s"%("Continue",self.i,self.text)

class Begin_provider(Line):
  str="Begin_provider"
  def __init__(self,i,text,filename):
    Line.__init__(self,i,text,filename)
  def __repr__(self):
    return "%20s:%5d : %s"%("Begin_provider",self.i,self.text)

class Cont_provider(Line):
  str="Cont_provider"
  def __init__(self,i,text,filename):
    Line.__init__(self,i,text,filename)
  def __repr__(self):
    return "%20s:%5d : %s"%("Cont_provider",self.i,self.text)

class End_provider(Line):
  str="End_provider"
  def __init__(self,i,text,filename):
    Line.__init__(self,i,text,filename)
  def __repr__(self):
    return "%20s:%5d : %s"%("End_provider",self.i,self.text)

class Begin_doc(Line):
  str="Begin_doc"
  def __init__(self,i,text,filename):
    Line.__init__(self,i,text,filename)
  def __repr__(self):
    return "%20s:%5d : %s"%("Begin_doc",self.i,self.text)

class Doc(Line):
  str="Doc"
  def __init__(self,i,text,filename):
    Line.__init__(self,i,text,filename)
  def __repr__(self):
    return "%20s:%5d : %s"%("Doc",self.i,self.text)

class End_doc(Line):
  str="End_doc"
  def __init__(self,i,text,filename):
    Line.__init__(self,i,text,filename)
  def __repr__(self):
    return "%20s:%5d : %s"%("End_doc",self.i,self.text)

class Begin_shell(Line):
  str="Begin_shell"
  def __init__(self,i,text,filename):
    Line.__init__(self,i,text,filename)
  def __repr__(self):
    return "%20s:%5d : %s"%("Begin_shell",self.i,self.text)

class End_shell(Line):
  str="End_shell"
  def __init__(self,i,text,filename):
    Line.__init__(self,i,text,filename)
  def __repr__(self):
    return "%20s:%5d : %s"%("End_shell",self.i,self.text)

class Begin_template(Line):
  str="Begin_template"
  def __init__(self,i,text,filename):
    Line.__init__(self,i,text,filename)
  def __repr__(self):
    return "%20s:%5d : %s"%("Begin_template",self.i,self.text)

class End_template(Line):
  str="End_template"
  def __init__(self,i,text,filename):
    Line.__init__(self,i,text,filename)
  def __repr__(self):
    return "%20s:%5d : %s"%("End_template",self.i,self.text)

class Subst(Line):
  str="Subst"
  def __init__(self,i,text,filename):
    Line.__init__(self,i,text,filename)
  def __repr__(self):
    return "%20s:%5d : %s"%("Subst",self.i,self.text)

class Assert(Line):
  str="Assert"
  def __init__(self,i,text,filename):
    Line.__init__(self,i,text,filename)
  def __repr__(self):
    return "%20s:%5d : %s"%("Assert",self.i,self.text)

class Touch(Line):
  str="Touch"
  def __init__(self,i,text,filename):
    Line.__init__(self,i,text,filename)
  def __repr__(self):
    return "%20s:%5d : %s"%("Touch",self.i,self.text)

class SoftTouch(Touch):
  str="SoftTouch"
  def __init__(self,i,text,filename):
    Line.__init__(self,i,text,filename)
  def __repr__(self):
    return "%20s:%5d : %s"%("SoftTouch",self.i,self.text)

class Irp_read(Line):
  str="Irp_read"
  def __init__(self,i,text,filename):
    Line.__init__(self,i,text,filename)
  def __repr__(self):
    return "%20s:%5d : %s"%("Irp_read",self.i,self.text)

class Irp_write(Line):
  str="Irp_write"
  def __init__(self,i,text,filename):
    Line.__init__(self,i,text,filename)
  def __repr__(self):
    return "%20s:%5d : %s"%("Irp_write",self.i,self.text)

class Irp_If(Line):
  str="Irp_If"
  def __init__(self,i,text,filename):
    Line.__init__(self,i,text,filename)
  def __repr__(self):
    return "%20s:%5d : %s"%("Irp_If",self.i,self.text)

class Irp_Else(Line):
  str="Irp_Else"
  def __init__(self,i,text,filename):
    Line.__init__(self,i,text,filename)
  def __repr__(self):
    return "%20s:%5d : %s"%("Irp_Else",self.i,self.text)

class Irp_Endif(Line):
  str="Irp_Endif"
  def __init__(self,i,text,filename):
    Line.__init__(self,i,text,filename)
  def __repr__(self):
    return "%20s:%5d : %s"%("Irp_Endif",self.i,self.text)

class Openmp(Line):
  str="Openmp"
  def __init__(self,i,text,filename):
    Line.__init__(self,i,text,filename)
  def __repr__(self):
    return "%20s:%5d : %s"%("Openmp",self.i,self.text)

class Directive(Line):
  str="Directive"
  def __init__(self,i,text,filename):
    Line.__init__(self,i,text,filename)
  def __repr__(self):
    return "%20s:%5d : %s"%("Directive",self.i,self.text)

class Use(Line):
  str="Use"
  def __init__(self,i,text,filename):
    Line.__init__(self,i,text,filename)
  def __repr__(self):
    return "%20s:%5d : %s"%("Use",self.i,self.text)

class Do(Line):
  str="Do"
  def __init__(self,i,text,filename):
    Line.__init__(self,i,text,filename)
  def __repr__(self):
    return "%20s:%5d : %s"%("Do",self.i,self.text)

class Enddo (Line):
  str="Enddo"
  def __init__(self,i,text,filename):
    Line.__init__(self,i,text,filename)
  def __repr__(self):
    return "%20s:%5d : %s"%("Enddo",self.i,self.text)

class If(Line):
  str="If"
  def __init__(self,i,text,filename):
    Line.__init__(self,i,text,filename)
  def __repr__(self):
    return "%20s:%5d : %s"%("If",self.i,self.text)

class Elseif(Line):
  str="Elseif"
  def __init__(self,i,text,filename):
    Line.__init__(self,i,text,filename)
  def __repr__(self):
    return "%20s:%5d : %s"%("Elseif",self.i,self.text)

class Else(Line):
  str="Else"
  def __init__(self,i,text,filename):
    Line.__init__(self,i,text,filename)
  def __repr__(self):
    return "%20s:%5d : %s"%("Else",self.i,self.text)

class Endif(Line):
  str="Endif"
  def __init__(self,i,text,filename):
    Line.__init__(self,i,text,filename)
  def __repr__(self):
    return "%20s:%5d : %s"%("Endif",self.i,self.text)

class Select(Line):
  str="Select"
  def __init__(self,i,text,filename):
    Line.__init__(self,i,text,filename)
  def __repr__(self):
    return "%20s:%5d : %s"%("Select",self.i,self.text)

class Case(Line):
  str="Case"
  def __init__(self,i,text,filename):
    Line.__init__(self,i,text,filename)
  def __repr__(self):
    return "%20s:%5d : %s"%("Case",self.i,self.text)

class End_select(Line):
  str="End_select"
  def __init__(self,i,text,filename):
    Line.__init__(self,i,text,filename)
  def __repr__(self):
    return "%20s:%5d : %s"%("End_select",self.i,self.text)

class Program(Line):
  str="Program"
  def __init__(self,i,text,filename):
    Line.__init__(self,i,text,filename)
  def __repr__(self):
    return "%20s:%5d : %s"%("Program",self.i,self.text)

class Subroutine(Line):
  str="Subroutine"
  def __init__(self,i,text,filename):
    Line.__init__(self,i,text,filename)
  def __repr__(self):
    return "%20s:%5d : %s"%("Subroutine",self.i,self.text)

class Function(Line):
  str="Function"
  def __init__(self,i,text,filename):
    Line.__init__(self,i,text,filename)
  def __repr__(self):
    return "%20s:%5d : %s"%("Function",self.i,self.text)

class Call(Line):
  str="Call"
  def __init__(self,i,text,filename):
    Line.__init__(self,i,text,filename)
  def __repr__(self):
    return "%20s:%5d : %s"%("Call",self.i,self.text)

class Provide(Line):
  str="Provide"
  def __init__(self,i,text,filename):
    Line.__init__(self,i,text,filename)
  def __repr__(self):
    return "%20s:%5d : %s"%("Provide",self.i,self.text)

class NoDep(Line):
  str="NoDep"
  def __init__(self,i,text,filename):
    Line.__init__(self,i,text,filename)
  def __repr__(self):
    return "%20s:%5d : %s"%("NoDep",self.i,self.text)

class Return (Line):
  str="Return"
  def __init__(self,i,text,filename):
    Line.__init__(self,i,text,filename)
  def __repr__(self):
    return "%20s:%5d : %s"%("Return",self.i,self.text)

class Include(Line):
  str="Include"
  def __init__(self,i,text,filename):
    Line.__init__(self,i,text,filename)
  def __repr__(self):
    return "%20s:%5d : %s"%("Include",self.i,self.text)

class Implicit (Line):
  str="Implicit"
  def __init__(self,i,text,filename):
    Line.__init__(self,i,text,filename)
  def __repr__(self):
    return "%20s:%5d : %s"%("Implicit",self.i,self.text)

class Free(Line):
  str="Free"
  def __init__(self,i,text,filename):
    Line.__init__(self,i,text,filename)
  def __repr__(self):
    return "%20s:%5d : %s"%("Free",self.i,self.text)

class End(Line):
  str="End"
  def __init__(self,i,text,filename):
    Line.__init__(self,i,text,filename)
  def __repr__(self):
    return "%20s:%5d : %s"%("End",self.i,self.text)

class Provide_all (Line):
  str="Provide_all"
  def __init__(self,i,text,filename):
    Line.__init__(self,i,text,filename)
  def __repr__(self):
    return "%20s:%5d : %s"%("Provide_all",self.i,self.text)


######################################################################

def create_irpf90_files():
  result = []
  from command_line import command_line
  import os
  def is_irpf90_file(filename):
    return filename.endswith(".irp.f") and not filename.startswith('.')
  result = filter ( is_irpf90_file, os.listdir(os.getcwd()) )
  for dir in command_line.include_dir:
    try:
      os.stat(dir)
      result += map(lambda x: dir+x, filter ( is_irpf90_file, os.listdir(dir) ) )
    except:
      continue
  if command_line.do_codelet:
    result += [command_line.codelet[3]]
  return result
irpf90_files = create_irpf90_files()







#!/usr/bin/env python2
#   IRPF90 is a Fortran90 preprocessor written in Python for programming using
#   the Implicit Reference to Parameters (IRP) method.
#   Copyright (C) 2009 Anthony SCEMAMA 
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License along
#   with this program; if not, write to the Free Software Foundation, Inc.,
#   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
#   Anthony Scemama
#   LCPQ - IRSAMC - CNRS
#   Universite Paul Sabatier
#   118, route de Narbonne      
#   31062 Toulouse Cedex 4      
#   scemama@irsamc.ups-tlse.fr


from util import *
from subroutine import *
from variables import variables
from variable  import Variable
from irpf90_t import *

def create_subroutines():
  from preprocessed_text import preprocessed_text
  result = {}
  for filename, text in preprocessed_text:
    buffer = []
    inside = False
    for line in text:
      if type(line) in [ Subroutine, Function ]:
        inside = True
      if inside:
        buffer.append(line)
      if isinstance(line,End):
        if inside:
         v = Sub(buffer)
         result[v.name] = v
         buffer = []
        inside = False
  return result

def create_called_by(subs,vars):
  for s in subs.values() + vars.values():
    if type(s) == Variable and s.same_as != s.name:
        continue
    for x in s.calls:
      try:
        subs[x].called_by.append(s.name)
      except KeyError:
        pass

  for s in subs.values():
    s.called_by = make_single(s.called_by)
    s.called_by.sort()
    
subroutines = create_subroutines()
create_called_by(subroutines,variables)

if __name__ == '__main__':
  for v in subroutines.keys():
    print v



#!/usr/bin/env python2
#   IRPF90 is a Fortran90 preprocessor written in Python for programming using
#   the Implicit Reference to Parameters (IRP) method.
#   Copyright (C) 2009 Anthony SCEMAMA 
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License along
#   with this program; if not, write to the Free Software Foundation, Inc.,
#   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
#   Anthony Scemama
#   LCPQ - IRSAMC - CNRS
#   Universite Paul Sabatier
#   118, route de Narbonne      
#   31062 Toulouse Cedex 4      
#   scemama@irsamc.ups-tlse.fr


import sys
import re

LENMAX = 70
tabn = 2
tab = " "*tabn

class Grep(object):

  re_begin_program = re.compile(r"^\s*program\s",flags=re.I)
  def begin_program(self,string):
    return re.match(self.re_begin_program,string) is not None

  re_end_program = re.compile(r"\s*(end\s*!?$|end\s*program)",flags=re.I)
  def end_program(self,string):
    return re.match(self.re_end_program,string) is not None

  re_begin_subroutine = re.compile(r"^\s*(recursive)?\s*subroutine\s",flags=re.I)
  def begin_subroutine(self,string):
    return re.match(self.re_begin_subroutine,string) is not None

  re_end_subroutine = re.compile(r"\s*(end\s*!?$|end\s*subroutine)",flags=re.I)
  def end_subroutine(self,string):
    return re.match(self.re_end_subroutine,string) is not None

  re_begin_function = re.compile(r"^.*function\s+.*\(",flags=re.I)
  def begin_function(self,string):
    return re.match(self.re_begin_function,string) is not None

  re_end_function = re.compile(r"\s*(end\s*!?$|end\s*function)",flags=re.I)
  def end_function(self,string):
    return re.match(self.re_end_function,string) is not None

  re_begin_provider = re.compile(r"^\s*&?begin_provider\s",flags=re.I)
  def begin_provider(self,string):
    return re.match(self.re_begin_provider,string) is not None

  re_end_provider = re.compile(r"^\s*end_provider\s*(!.*)?$", flags=re.I)
  def end_provider(self,string):
    return re.match(self.re_end_provider,string) is not None

  re_begin_do = re.compile(r"^\s*do\s+",flags=re.I)
  def begin_do(self,string):
    return re.match(self.re_begin_do,string) is not None

  re_end_do = re.compile(r"^\s*end\s*do\s*(!.*)?$",flags=re.I)
  def end_do(self,string):
    return re.match(self.re_end_do,string) is not None

  re_begin_if = re.compile(r"^\s*if(\(|\s+).*(&|then)\s*(!.*)?$",flags=re.I)
  def begin_if(self,string):
    return re.match(self.re_begin_if,string) is not None

  re_else = re.compile(r"^\s*else",flags=re.I)
  def xelse(self,string):
    return re.match(self.re_else,string) is not None

  re_end_if = re.compile(r"^\s*end\s*if\s*(!.*)?$",flags=re.I)
  def end_if(self,string):
    return re.match(self.re_end_if,string) is not None

  re_begin_select = re.compile(r"^\s*select\s*case",flags=re.I)
  def begin_select(self,string):
    return re.match(self.re_begin_select,string) is not None

  re_case = re.compile(r"^\s*case\s*\(",flags=re.I)
  def case(self,string):
    return re.match(self.re_case,string) is not None

  re_end_select = re.compile(r"^\s*end\s*select\s*(!.*)?$",flags=re.I)
  def end_select(self,string):
    return re.match(self.re_end_select,string) is not None

  re_continuation = re.compile(r"^\s*\S+.*&")
  def continuation(self,string):
    return re.match(self.re_continuation,string) is not None

  re_declaration = re.compile(r"^.*::.*$")
  def declaration(self,string):
    return re.match(self.re_declaration,string) is not None

grep = Grep()

class indent(object):

  def __init__(self):
    """Run the program"""
    self.run()

  def format_declaration(self,string,n):
    l,r = string.split('::')
    return l.strip().ljust(n) + ' :: '+ r.strip()

  def format_continuation(self,string,n):
    buffer = string.split('&')
    if len(buffer) == 1:
      l = buffer[0]
      return l
    else:
      l, r = buffer
    return l.strip().ljust(69-len(n)) + '&'+ r.strip()

  def get_filename(self):
    """The file name is the first argument"""
    if '_filename' not in self.__dict__:
      try:
        self._filename = sys.argv[1]
      except:
        self._filename = None
    return self._filename
  filename=property(fget=get_filename)

  def get_text(self):
    """The text of the file is a list of lines"""
    if '_text' not in self.__dict__:
      if self.filename is not None:
        f = open(self.filename,'r')
        self._text = f.read().splitlines()
        f.close()
      else:
        self._text = sys.stdin.read().splitlines()
    return self._text
  text=property(fget=get_text)

  def indentlevel(self,line):
    line = line.rstrip()
    k=0
    if len(line) > 0:
      while line[k] == ' ':
        k+=1
    return k

  def run(self):
    lines = self.text
    indent0 = " "*self.indentlevel(self.text[0])
    k = indent0
    line = ""
    for i in range(len(self.text)):
      prevline = line
      line = self.text[i].strip()
      if grep.continuation(line):
        line = self.format_continuation(line,k)

      if grep.continuation(prevline):
        print k+2*tab+self.format_continuation(line,k+2*tab)
        continue

      if grep.begin_subroutine(line):
        print line
        k = indent0+tab
        continue

      if grep.begin_function(line):
        print line
        k = indent0+tab
        continue

      if grep.begin_program(line):
        print line
        k = indent0+tab
        continue

      if grep.begin_provider(line):
        if line[0] != '&':
          k = indent0+tab
          if grep.begin_provider(self.text[i+1].strip()):
            print " "+line
          else:
            print line
        else:
          print line
        continue

      if grep.declaration(line):
        print k+self.format_declaration(line,30)
        continue

      if grep.begin_do(line):
        print k+line
        k += tab
        continue

      if grep.begin_if(line):
        print k+line
        k += tab
        continue

      if grep.xelse(line):
        print k[:-tabn]+line
        continue

      if grep.begin_select(line):
        print k+line
        k += 2*tab
        continue

      if grep.case(line):
        print k[:-tabn]+line
        continue

      if grep.end_do(line):
        k = k[:-tabn]
        print k+line
        continue

      if grep.end_if(line):
        k = k[:-tabn]
        print k+line
        continue

      if grep.end_select(line):
        k = k[:-2*tabn]
        print k+line
        continue

      if grep.end_subroutine(line):
        print line
        k = indent0
        continue

      if grep.end_function(line):
        print line
        k = indent0
        continue

      if grep.end_provider(line):
        print line
        k = indent0
        continue

      if grep.end_program(line):
        print line
        k = indent0
        continue

      print k+line


def main():
  indent()


if __name__ == '__main__':
  main()




#!/usr/bin/env python2
#   IRPF90 is a Fortran90 preprocessor written in Python for programming using
#   the Implicit Reference to Parameters (IRP) method.
#   Copyright (C) 2009 Anthony SCEMAMA 
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License along
#   with this program; if not, write to the Free Software Foundation, Inc.,
#   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
#   Anthony Scemama
#   LCPQ - IRSAMC - CNRS
#   Universite Paul Sabatier
#   118, route de Narbonne      
#   31062 Toulouse Cedex 4      
#   scemama@irsamc.ups-tlse.fr


from variable  import Variable
from variables import variables
from subroutine  import Sub
from subroutines import subroutines
from irpf90_t  import *
from util import *


def do_print_short(file,var):
  """Makes a short print, as in irpf90_entities"""
  assert type(var) == Variable
  print >>file, "%s : %s :: %s %s"%( \
   var.line.filename[0].ljust(35),
   var.type.ljust(30),
   var.name.ljust(25),
   build_dim(var.dim) )

######################################################################
def process_doc(file,line):
  assert type(line) == str
  line = line.strip()
  if line == "":
    line = ".br"
  print >>file, line

######################################################################
def process_deps(file,l):
  assert type(l) == list
  for v in l:
    print >>file, "%s\n.br"%(v,)

######################################################################
def process_types(file,var):
  assert type(var) == Variable
  vars = [var.name] + var.others
  for var in vars:
    name = var
    var = variables[var]
    Type = var.type
    dim = build_dim(var.dim)
    print >>file, "%s\t:: %s\t%s"%(Type,name,dim)

######################################################################
def do_print(var):
  assert type(var) == Variable
  filename = var.line.filename[0]
  name = var.name
  file = open("%s%s.l"%(mandir,var.name), "w")
  print >>file, '.TH "IRPF90 entities" l %s "IRPF90 entities" %s'%(name,name)
  if var.same_as != var.name:
    var = variables[var.same_as]
  print >>file, ".SH Declaration"
  print >>file, ".nf"
  process_types(file,var)
  print >>file, ".ni"
  if var.doc != []:
   print >>file, ".SH Description"
   for l in var.doc:
     process_doc(file,l)
  print >>file, ".SH File\n.P"
  print >>file, filename
  if var.needs != []:
    var.needs.sort()
    print >>file, ".SH Needs"
    process_deps(file,var.needs)
  if var.needed_by != []:
    var.needed_by.sort()
    print >>file, ".SH Needed by"
    process_deps(file,var.needed_by)
  print >>file, ".SH Instability factor"
  fo = len(var.children)
  fi = len(var.parents)
  print >>file, "%5.1f %%"%(100.* (fi / (fi+fo+.000001) ))
  print >>file, ".br"
  file.close()

######################################################################
def do_print_rst(var):
  """Print providers in a format suitable for sphinx"""
  assert type(var) == Variable
  filename = var.line.filename[0]
  name = var.name
  file = open("%s%s.rst"%(mandir,var.name), "w")
  print >>file, ".. c:var:: %s\n"%(var.name.lower())
  print >>file, ""
  print >>file, "    File : :file:`"+filename+"`"
  print >>file, ""
  print >>file, "    .. code:: fortran"
  print >>file, ""
  if var.same_as != var.name:
    var = variables[var.same_as]
  for v in [var.name] + var.others:
    name = v
    v = variables[v]
    Type = v.type
    dim = build_dim(v.dim)
    print >>file, "        %s\t:: %s\t%s"%(Type,name,dim)
  print >>file, ""
  print >>file, ""

  if var.doc != []:
    d = []
    for text in var.doc:
        text_old = None
        while text_old != text:
            text_old = text
            text = text.replace("$",":math:`",1).replace("$","` ",1)
        d.append(text)
    loop = True
    while loop:
      maxlen=0
      for l in d:
        ll = len(l)
        maxlen = max(ll,maxlen)
        if ll > 0 and l[0] != ' ':
          loop = False
          break
      loop = loop and maxlen > 0
      if loop:
        d = [ l[1:] for l in d ]
    for line in d:
        print >>file, "    "+line
  print >>file, ""
  if var.needs != []:
    var.needs.sort()
    print >>file, "    Needs:"
    print >>file, ""
    print >>file, "    .. hlist::"
    print >>file, "       :columns: 3"
    print >>file, ""
    for v in var.needs:
        print >>file, "       * :c:data:`%s`"%(variables[v].same_as.lower(),)
    print >>file, ""
  if var.needed_by != []:
    var.needed_by.sort()
    print >>file, "    Needed by:"
    print >>file, ""
    print >>file, "    .. hlist::"
    print >>file, "       :columns: 3"
    print >>file, ""
    for v in var.needed_by:
        print >>file, "       * :c:data:`%s`"%(variables[v].same_as.lower(),)
  print >>file, ""
  file.close()

######################################################################
def process_declaration_subroutine(file, sub):
  print >>file, sub.line.text.split('!')[0].strip()

# for line in sub.text:
######################################################################
def do_print_subroutines(sub):
  assert type(sub) == Sub
  filename = sub.line.filename
  name = sub.name
  file = open("%s%s.l"%(mandir,sub.name), "w")
  print >>file, '.TH "IRPF90 entities" l %s "IRPF90 entities" %s'%(name,name)
  print >>file, ".SH Declaration"
  print >>file, ".nf"
  process_declaration_subroutine(file,sub)
  print >>file, ".ni"
  if sub.doc != []:
   print >>file, ".SH Description"
   for l in sub.doc:
     process_doc(file,l)
  print >>file, ".SH File\n.P"
  print >>file, filename
  if sub.needs != []:
    sub.needs.sort()
    print >>file, ".SH Needs"
    process_deps(file,sub.needs)
  if sub.called_by != []:
    sub.called_by.sort()
    print >>file, ".SH Called by"
    process_deps(file,sub.called_by)
  if sub.calls != []:
    sub.calls.sort()
    print >>file, ".SH Calls"
    process_deps(file,sub.calls)
  if sub.touches != []:
    sub.touches.sort()
    print >>file, ".SH Touches"
    process_deps(file,sub.touches)
  print >>file, ".SH Instability factor"
  fo = len(sub.needs)+len(sub.calls)+len(sub.touches)
  fi = len(sub.called_by)
  print >>file, "%5.1f %%"%(100.* (fi / (fi+fo+.000001) ))
  print >>file, ".br"
  file.close()

######################################################################
def do_print_subroutines_rst(sub):
  """Print subroutines in a format suitable for sphinx"""
  assert type(sub) == Sub
  filename = sub.line.filename
  name = sub.name
  file = open("%s%s.rst"%(mandir,sub.name), "w")
  print >>file, ".. c:function:: %s:\n"%(sub.name.lower())
  print >>file, ""
  print >>file, "    File : :file:`"+filename+"`"
  print >>file, ""
  print >>file, "    .. code:: fortran"
  print >>file, ""
  print >>file, "        "+sub.line.text.split('!')[0].strip()
  print >>file, ""
  print >>file, ""
  if sub.doc != []:
    d = list(sub.doc)
    loop = True
    while loop:
      maxlen=0
      for l in d:
        ll = len(l)
        maxlen = max(ll,maxlen)
        if ll > 0 and l[0] != ' ':
          loop = False
          break
      loop = loop and maxlen > 0
      if loop:
        d = [ l[1:] for l in d ]
    for l in d:
        print >>file, "    "+l
  print >>file, ""
  if sub.needs != []:
    sub.needs.sort()
    print >>file, "    Needs:"
    print >>file, ""
    print >>file, "    .. hlist::"
    print >>file, "       :columns: 3"
    print >>file, ""
    for v in sub.needs:
        print >>file, "       * :c:data:`%s`"%(variables[v].same_as.lower(),)
    print >>file, ""
  if sub.called_by != []:
    sub.called_by.sort()
    print >>file, "    Called by:"
    print >>file, ""
    print >>file, "    .. hlist::"
    print >>file, "       :columns: 3"
    print >>file, ""
    for v in sub.called_by:
        if v in subroutines:
            print >>file, "       * :c:func:`%s`"%(v.lower(),)
        elif v in variables:
            print >>file, "       * :c:data:`%s`"%(variables[v.lower()].same_as.lower(),)
    print >>file, ""
  if sub.calls != []:
    sub.calls.sort()
    print >>file, "    Calls:"
    print >>file, ""
    print >>file, "    .. hlist::"
    print >>file, "       :columns: 3"
    print >>file, ""
    for v in sub.calls:
        print >>file, "       * :c:func:`%s`"%(v.lower(),)
    print >>file, ""
  if sub.touches != []:
    sub.touches.sort()
    print >>file, "    Touches:"
    print >>file, ""
    print >>file, "    .. hlist::"
    print >>file, "       :columns: 3"
    print >>file, ""
    for v in sub.touches:
        print >>file, "       * :c:data:`%s`"%(variables[v.lower()].same_as.lower(),)
    print >>file, ""
  file.close()

######################################################################
def run():
  import parsed_text
  import os,sys
  pid1 = os.fork()
  if pid1 == 0:
    for v in variables.values():
      do_print(v)
      do_print_rst(v)
    for s in subroutines.values():
      do_print_subroutines(s)
      do_print_subroutines_rst(s)
    sys.exit(0)

  pid2 = os.fork()
  if pid2 == 0:
    tags = []
    l = variables.keys()
    file = open("irpf90_entities","w")
    l.sort()
    for v in l:
      do_print_short(file,variables[v])
      line = variables[v].line
      tags.append( '%s\t%s\t%d\n'%(v,line.filename[0],line.i) )
    file.close()
    l = subroutines.keys()
    for v in l:
      line = subroutines[v].line
      tags.append('%s\t%s\t%d\n'%(v,line.filename,line.i))
    tags.sort()
    file = open("tags","w")
    for line in tags:
      file.write(line)
    file.close()
    sys.exit(0)

  os.waitpid(pid1,0)
  os.waitpid(pid2,0)

######################################################################
if __name__ == '__main__':
  run()



#!/usr/bin/env python2
#   IRPF90 is a Fortran90 preprocessor written in Python for programming using
#   the Implicit Reference to Parameters (IRP) method.
#   Copyright (C) 2009 Anthony SCEMAMA 
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License along
#   with this program; if not, write to the Free Software Foundation, Inc.,
#   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
#   Anthony Scemama
#   LCPQ - IRSAMC - CNRS
#   Universite Paul Sabatier
#   118, route de Narbonne      
#   31062 Toulouse Cedex 4      
#   scemama@irsamc.ups-tlse.fr


import getopt, sys
from version import version
import re

description = "IRPF90 Fortran preprocessor."
options = {}
options['a'] = [ 'assert'       , 'Activates ASSERT statements. If absent, remove ASSERT statements.', 0 ]
options['c'] = [ 'codelet'      , 'entity:NMAX or entity:precondition:NMAX  : Generate a codelet to profile a provider running NMAX times', 1 ]
options['C'] = [ 'coarray'      , 'All providers are coarrays', 0 ]
options['d'] = [ 'debug'        , 'Activates debug. The name of the current subroutine/function/provider will be printed on the standard output when entering or exiting a routine, as well as the CPU time passed inside the routine.', 0 ]
options['D'] = [ 'define'       , 'Defines a variable identified by the IRP_IF statements.', 1 ]
options['g'] = [ 'profile'      , 'Activates profiling of the code.', 0 ]
options['h'] = [ 'help'         , 'Print this help', 0 ]
options['I'] = [ 'include'      , 'Include directory', 1 ]
options['j'] = [ 'ninja'        , 'Use Ninja instead of make', 0 ]
options['i'] = [ 'init'         , 'Initialize current directory. Creates a default Makefile and the temporary working directories.', 0 ]
options['l'] = [ 'align'        , 'Align arrays using compiler directives and sets the $IRP_ALIGN variable. For example, --align=32 aligns all arrays on a 32 byte boundary.', 1 ]
options['m'] = [ 'memory'       , 'Print memory allocations/deallocations.', 0 ]
options['n'] = [ 'inline'       , '<all|providers|builders> : Force inlining of providers or builders', 1 ]
options['o'] = [ 'checkopt'     , 'Shows where optimization may be required', 0 ]
options['p'] = [ 'preprocess'   , 'Prints a preprocessed file to standard output. Useful for  debugging files containing shell scripts.', 1 ]
options['r'] = [ 'no_directives', 'Ignore all compiler directives !DEC$ and !DIR$', 0 ]
options['s'] = [ 'substitute'   , 'Substitute values in do loops for generating specific optimized code.', 1 ]
options['t'] = [ 'touch'        , 'Display which entities are touched when touching the variable given as an argument.', 1 ]
options['v'] = [ 'version'      , 'Prints version of irpf90', 0 ]
options['w'] = [ 'warnings'     , 'Activate Warnings', 0 ]
options['z'] = [ 'openmp'       , 'Activate for OpenMP code', 0 ]

class CommandLine(object):

  def __init__(self):
    global options
    self._opts = None
    self.argv = list(sys.argv)
    self.executable_name = self.argv[0]

  def defined(self):
    if '_defined' not in self.__dict__:
      self._defined = []
      for o,a in self.opts:
        if o in [ "-D", '--'+options['D'][0] ]:
          self._defined.append(a)
    return self._defined
  defined = property(fget=defined)

  def include_dir(self):
    if '_include_dir' not in self.__dict__:
      self._include_dir = []
      for o,a in self.opts:
        if o in [ "-I", '--'+options['I'][0] ]:
          if len(a) < 1: 
            print "Error: -I option needs a directory"
          if a[-1] != '/':
            a = a+'/'
          self._include_dir.append(a)
    return self._include_dir
  include_dir = property(fget=include_dir)

  def inline(self):
    if '_inline' not in self.__dict__:
      self._inline = ""
      for o,a in self.opts:
        if o in [ "-n", '--'+options['n'][0] ]:
          self._inline = a
          break
    return self._inline
  inline = property(fget=inline)

  def substituted(self):
    if '_substituted' not in self.__dict__:
      self._substituted = {}
      for o,a in self.opts:
        if o in [ "-s", '--'+options['s'][0] ]:
          k, v = a.split(':')
          v_re = re.compile(r"(\W)(%s)(\W.*$|$)"%k.strip())
          self._substituted[k] = [v, v_re]
    return self._substituted
  substituted = property(fget=substituted)

  def codelet(self):
    if '_codelet' not in self.__dict__:
      self._codelet = []
      for o,a in self.opts:
        if o in [ "-c", '--'+options['c'][0] ]:
          buffer = a.split(':')
          filename = 'codelet_'+buffer[0]+'.irp.f'
          if len(buffer) == 2:
            self._codelet = [buffer[0], int(buffer[1]), None, filename]
          elif len(buffer) == 3:
            self._codelet = [buffer[0], int(buffer[2]), buffer[1], filename]
          else:
            print """
Error in codelet definition. Use:
--codelet=provider:NMAX
or
--codelet=provider:precondition:NMAX
"""
            sys.exit(1)
    return self._codelet
  codelet = property(fget=codelet)

  def preprocessed(self):
    if '_preprocessed' not in self.__dict__:
      self._preprocessed = []
      for o,a in self.opts:
        if o in [ "-p", '--'+options['p'][0] ]:
          self._preprocessed.append(a)
    return self._preprocessed
  preprocessed = property(fget=preprocessed)

  def touched(self):
    if '_touched' not in self.__dict__:
      self._touched = []
      for o,a in self.opts:
        if o in [ "-t", '--'+options['t'][0] ]:
          self._touched.append(a.lower())
    return self._touched
  touched = property(fget=touched)

  def align(self):
    if '_align' not in self.__dict__:
      self._align = '1'
      for o,a in self.opts:
        if o in [ "-l", '--'+options['l'][0] ]:
          self._align = a
    return self._align
  align = property(fget=align)

  def coarray(self):
    if '_coarray' not in self.__dict__:
      self._coarray = False
      for o,a in self.opts:
        if o in [ "-C", '--'+options['C'][0] ]:
          self._coarray = True
    return self._coarray
  coarray = property(fget=coarray)

  def warnings(self):
    if '_warnings' not in self.__dict__:
      self._warnings= False
      for o,a in self.opts:
        if o in [ "-w", '--'+options['w'][0] ]:
          self._warnings= True
    return self._warnings
  do_warnings = property(fget=warnings)

  def openmp(self):
    if '_openmp' not in self.__dict__:
      self._openmp = False
      for o,a in self.opts:
        if o in [ "-z", '--'+options['z'][0] ]:
          self._openmp = True
    return self._openmp
  do_openmp = property(fget=openmp)

  def ninja(self):
    if '_ninja' not in self.__dict__:
      self._ninja = False
      for o,a in self.opts:
        if o in [ "-j", '--'+options['j'][0] ]:
          self._ninja = True
    return self._ninja
  do_ninja = property(fget=ninja)

  def directives(self):
    if '_directives' not in self.__dict__:
      self._directives = True
      for o,a in self.opts:
        if o in [ "-r", '--'+options['r'][0] ]:
          self._directives = False
    return self._directives
  directives = property(fget=directives)

  def usage(self):
    t = """
$EXE - $DESCR

Usage:
  $EXE [OPTION]

Options:
"""
    t = t.replace("$EXE",self.executable_name)
    t = t.replace("$DESCR",description)
    print t
    print_options()
    print ""
    print "Version : ", version
    print ""

  def opts(self):
    if self._opts is None:
      optlist = ["",[]]
      for o in options:
        b = [o]+options[o]
        if b[3] == 1:
          b[0] = b[0]+":"
          b[1] = b[1]+"="
        optlist[0] += b[0]
        optlist[1] += [b[1]]
    
      try:
        self._opts, args = getopt.getopt(self.argv[1:], optlist[0], optlist[1])
      except getopt.GetoptError, err:
        # print help information and exit:
        self.usage()
        print str(err) # will print something like "option -a not recognized"
        sys.exit(2)
    
    return self._opts
  opts = property(fget=opts)
  
  t = """
def do_$LONG(self):
    if '_do_$LONG' not in self.__dict__:
      self._do_$LONG = False
      for o,a in self.opts:
        if o in ("-$SHORT", "--$LONG"):
          self._do_$LONG = True
          break
    return self._do_$LONG
do_$LONG = property(fget=do_$LONG)
"""
  for short in options:
    long = options[short][0]
    exec t.replace("$LONG",long).replace("$SHORT",short) #in locals()

  def do_run(self):
   if '_do_run' not in self.__dict__:
     self._do_run = not ( \
       self.do_version or \
       self.do_help    or \
       self.do_preprocess or \
       self.do_touch or \
       self.do_init )
   return self._do_run
  do_run = property(fget=do_run)


command_line = CommandLine()

def print_options():
  keys = options.keys()
  keys.sort()
  import subprocess
  for k in keys:
    description = options[k][1]
    p1 = subprocess.Popen(["fold", "-s", "-w", "40"],stdout=subprocess.PIPE,stdin=subprocess.PIPE)
    description = p1.communicate(description)[0]
    description = description.replace('\n','\n'.ljust(27))
    print ("-%s, --%s"%(k,options[k][0])).ljust(25), description+'\n'
  print "\n"

if __name__ == '__main__':
  print_options()



#!/usr/bin/env python2
#   IRPF90 is a Fortran90 preprocessor written in Python for programming using
#   the Implicit Reference to Parameters (IRP) method.
#   Copyright (C) 2009 Anthony SCEMAMA
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License along
#   with this program; if not, write to the Free Software Foundation, Inc.,
#   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
#   Anthony Scemama
#   LCPQ - IRSAMC - CNRS
#   Universite Paul Sabatier
#   118, route de Narbonne
#   31062 Toulouse Cedex 4
#   scemama@irsamc.ups-tlse.fr

from irpf90_t import *
from regexps import *
import error
from command_line import command_line
from util import *

# Local regular expressions
re_endif = re.compile("end\s+if")
re_elseif = re.compile("else\s+if")
re_enddo  = re.compile("end\s+do")
re_endtype= re.compile("end\s+type")
re_endmodule = re.compile("end\s+module")
re_endselect  = re.compile("end\s+select")
re_endinterface  = re.compile("end\s+interface")

# Local variables
Free_form = 0
Fixed_form = 1

######################################################################
# Dictionary of simple statements
simple_dict = {
        "subroutine":            Subroutine      ,
        "begin_shell":           Begin_shell     ,
        "end_shell":             End_shell       ,
        "begin_template":        Begin_template  ,
        "end_template":          End_template    ,
        "subst":                 Subst           ,
        "end_doc":               End_doc         ,
        "begin_provider":        Begin_provider  ,
        "&begin_provider":       Cont_provider   ,
        "end_provider":          End_provider    ,
        "assert":                Assert          ,
        "touch":                 Touch           ,
        "soft_touch":            SoftTouch       ,
        "provide":               Provide         ,
        "no_dep":                NoDep           ,
        "free":                  Free            ,
        "irp_if":                Irp_If          ,
        "irp_else":              Irp_Else        ,
        "irp_endif":             Irp_Endif       ,
        "irp_read":              Irp_read        ,
        "irp_write":             Irp_write       ,
        "use":                   Use             ,
        "do":                    Do              ,
        "if":                    If              ,
        "case":                  Case            ,
        "elseif":                Elseif          ,
        "else":                  Else            ,
        "enddo":                 Enddo           ,
        "endif":                 Endif           ,
        "endselect":             End_select      ,
        "end":                   End             ,
        "include":               Include         ,
        "call":                  Call            ,
        "continue":              Continue        ,
        "return":                Return          ,
        "implicit":              Implicit        ,
        "save":                  Declaration     ,
        "function":              Function        ,
        "recursive":             Function        ,
}

def get_type (i, filename, line, is_doc):
  '''Find the type of a text line'''
  assert type(i) == int
  assert type(filename) == str
  assert type(line) == str
  assert type(is_doc) == bool

  # add support for interface block
  has_interface = False

  line = line.rstrip()
  line = line.replace("$IRP_ALIGN",command_line.align)
  lower_line0 = line.lstrip().lower()
  lower_line = lower_line0.replace("!"," ! ")

  # Replacements
  lower_line = re_elseif.sub("elseif",lower_line)
  lower_line = re_enddo.sub("enddo",lower_line)
  lower_line = re_endtype.sub("endtype",lower_line)
  lower_line = re_endmodule.sub("endmodule",lower_line)
  lower_line = re_endif.sub("endif",lower_line)
  lower_line = re_endselect.sub("endselect",lower_line)
  lower_line = re_endinterface.sub("endinterface",lower_line)

  for c in """()'"[]""":
    lower_line = lower_line.replace(c," "+c+" ")

  buffer = lower_line.split()
  if len(buffer) == 0:
    return [ Empty_line(i,line,filename) ], is_doc

  firstword = buffer[0]
  if firstword.isdigit():
    assert len(buffer) > 1
    buffer = buffer[1:]
    firstword = buffer[0]

  # Identify line
  if firstword == "endinterface":
    has_interface = False
    return [ Simple_line   (i,line,filename) ], False

  if firstword == "interface" or has_interface:
    has_interface = True
    return [ Simple_line   (i,line,filename) ], False

  if firstword == "end_doc":
    return [ End_doc       (i,line,filename) ], False

  if firstword == "begin_doc":
    return [ Begin_doc       (i,line,filename) ], True

  if firstword == "program":
    return [ Program         (i,line,filename) ], False

  if is_doc:
    return [ Doc (i,line,filename) ], is_doc

  if firstword in simple_dict:
    return [ simple_dict[firstword](i,line,filename) ], is_doc

  if firstword in [ "select", "selectcase" ]:
    return [ Select(i,line,filename) ] , is_doc

  if len(lower_line0) > 4:

    if firstword[0] == '#':
      result = [ Simple_line(i,line,filename) ]
      error.warn ( result[0] ,
"""irpf90 may not work with preprocessor directives. You can use
 Irp_if ... Irp_else ... Irp_endif
instead of
 #ifdef ... #else ... #endif""" )
      return result, is_doc

    if firstword.startswith("case("):
     return [ Case(i,line,filename) ], is_doc

    if lower_line0[1:5] == "$omp":
     return [ Openmp(i,line,filename) ], is_doc
    elif lower_line0[1:5] in ["dec$", "dir$"] and command_line.directives:
     return [ Directive(i,line,filename) ], is_doc
    elif lower_line0[1:3] == "$ ":
     return [ Openmp(i,line,filename) ], is_doc

    if re_decl.match(lower_line) is not None:
      if "function" in buffer[1:3]:
        return [ Function (i,line,filename) ], is_doc
      else:
        return [ Declaration (i,line,filename) ], is_doc

    # Detect errors
    if firstword == "dowhile":
      error.fail( Do(i,line,filename) , "'do while' should be in 2 words." )

  return [ Simple_line(i,line,filename) ], is_doc


######################################################################
def get_text(lines,filename):
  '''Read the input file and transform it to labeled lines'''
  assert type(filename) == str
  assert type(lines) == list

  result = []
  is_doc = False
  for i,line in enumerate(lines):
    line, is_doc = get_type(i+1,filename,line,is_doc)
    result += line
  return result

######################################################################
def execute_shell(text):
  '''Execute the embedded shell scripts'''
  def fail(l,a,b): error.fail(l,"In Begin_Shell, %s '%s'"%(a,b))
  inside = False
  result = []
  for line in text:
    if inside:
      if type(line) == Begin_shell:
        error.fail(line,"Nested Begin_shell")
      elif type(line) == End_shell:
        inside = False
        # Write script file
        scriptname = "%s%s_shell_%d"%(irpdir,line.filename,line.i)
        file = open(scriptname,'w')
        file.writelines(script)
        file.close()
        scriptname = "%s_shell_%d"%(line.filename,line.i)
        file = open(scriptname,'w')
        file.writelines(script)
        file.close()
        # Execute shell
        import os
        pipe = os.popen("%s < %s"%(shell,scriptname),'r')
        lines = pipe.readlines()
        pipe.close()
        result += get_text(lines,scriptname)
        os.remove(scriptname)
      else:
        script.append(line.text+'\n')
    else:
      if type(line) == Begin_shell:
        inside = True
        begin = line.i
        script = []
        # Find shell executable
        buffer = line.text.split('[')
        if len(buffer) > 2:
          fail(line,"Too many",'[')
        elif len(buffer) < 2:
          fail(line,"Missing",'[')
        buffer = buffer[1]
        buffer = buffer.split(']')
        if len(buffer) > 2:
          fail(line,"Too many",']')
        elif len(buffer) < 2:
          fail(line,"Missing",']')
        shell = buffer[0].strip()
      elif type(line) == End_shell:
        error.fail(line,"Begin_shell missing")
      else:
        result.append(line)
  return result

######################################################################
def execute_templates(text):
  '''Execute the templates'''
  def fail(l,a,b): error.fail(l,"In %s, %s"%(a,b))

  def get_variables(line):
    buffer = line.text.split('[',1)
    if len(buffer)<2:
      fail(line,"Subst","Syntax error")
    buffer = buffer[1].replace(']','')
    buffer = buffer.split(',')
    return map(lambda x: '$%s'%(x.strip()), buffer)

  TEMPLATE = 1
  SUBST = 2
  inside = 0
  result = []
  for line in text:
    if inside == 0:
      if type(line) == Begin_template:
        script = []
        inside = TEMPLATE
        script = "template = \"\"\"\n"
      else:
        result.append(line)
    elif inside == TEMPLATE:
      if type(line) == Begin_template:
        fail(line,"template", "Nested Begin_Template")
      elif type(line) == End_template:
        fail(line,"template","Missing Subst")
      elif type(line) == Subst:
        inside = SUBST
        script += "\"\"\"\n"
        variables = get_variables(line)
        script += "v = []\n"
        subst = ""
      else:
        script += line.text+"\n"
    else: # inside == SUBST
      if type(line) == Begin_template:
        fail(line,"subst","Nested Begin_template")
      elif type(line) == Subst:
        fail(line,"subst","Subst already defined")
      elif type(line) == End_template:
        inside = 0
        subst = subst.rstrip()
        if subst[-2:] == ';;':
          subst = subst[:-2]
        for s in subst.split(';;'):
          buffer = map(lambda x: x.strip(), s.split(';'))
          if len(buffer) != len(variables):
            fail(line,"subst","%d variables defined, and %d substitutions"%(len(variables),len(buffer)))
          script += "v.append( { \\\n"
          for t,v in zip(variables,buffer):
            script += ' "%s": """%s""" ,\n'%(t,v)
          script += "} )\n"
        script += "for d in v:\n  t0 = str(template)\n"
        for v in variables:
          script += "  t0 = t0.replace('%s',d['%s'])\n"%(v,v)
        script += "  print t0\n"
        # Write script file
        scriptname = "%s%s_template_%d"%(irpdir,line.filename,line.i)
        file = open(scriptname,'w')
        file.writelines(script)
        file.close()
        scriptname = "%s_template_%d"%(line.filename,line.i)
        file = open(scriptname,'w')
        file.writelines(script)
        file.close()
        # Execute shell
        import os
        pipe = os.popen("python2 < %s"%(scriptname),'r')
        lines = pipe.readlines()
        pipe.close()
        result += get_text(lines,scriptname)
        os.remove(scriptname)
      else:
        subst += line.text+'\n'

  return result

######################################################################
def form(text):
  '''Find if the text is in fixed form or in free form'''
  assert type(text) == list
  if len(text) == 0:
    return Free_form
  assert isinstance(text[0],Line)

  re2 = re.compile(r"^\s*[!#]")
  re3 = re.compile(r"^\s*[^ 0-9]+")
  for line in text:
    if type(line) in [ Empty_line, Doc, Openmp, Directive ]:
      pass
    else:
      if len(line.text) > 5:
        test = line.text[0:5]
        if test[0] in "Cc#!*":
          pass
        else:
          if re2.match(test) is None and \
             re3.match(test) is not None:
            return Free_form
          if line.text.rstrip()[-1] == '&':
            return Free_form
  return Fixed_form

######################################################################
def add_operators(text):
  re_incr = re.compile(r"(\s*)(.*)(\+=)(.*$)", re.S)
  re_decr = re.compile(r"(\s*)(.*)(-=)(.*$)", re.S)
  re_mult = re.compile(r"(\s*)(.*)(\*=)(.*$)", re.S)
  re_incr_if = re.compile(r"(.*)(\))(\s*)(.*)(\+=)(.*$)", re.S)
  re_decr_if = re.compile(r"(.*)(\))(\s*)(.*)(-=)(.*$)", re.S)
  re_mult_if = re.compile(r"(.*)(\))(\s*)(.*)(\*=)(.*$)", re.S)
  '''Change additional operators'''
  result = []
  for line in text:
    buffer = line.text
    ls = buffer.strip()
    if ls.startswith('print ') or \
       ls.startswith('print*') or \
       ls.startswith('write('):
        pass
    elif buffer.lstrip().startswith("if "):
        if "+=" in buffer:
            line.text = re.sub(re_incr_if, r'\1\2\4=\4+(\6)', buffer)
        elif "-=" in buffer:
            line.text = re.sub(re_incr_if, r'\1\2\4=\4-(\6)', buffer)
        elif "*=" in buffer:
            line.text = re.sub(re_incr_if, r'\1\2\4=\4*(\6)', buffer)
    else:
        if "+=" in buffer:
            line.text = re.sub(re_incr, r'\1\2=\2+(\4)', buffer)
        elif "-=" in buffer:
            line.text = re.sub(re_decr, r'\1\2=\2-(\4)', buffer)
        elif "*=" in buffer:
            line.text = re.sub(re_mult, r'\1\2=\2*(\4)', buffer)
    result.append(line)
  return result

######################################################################
def remove_comments(text,form):
  '''Remove all comments'''
  result = []

  def remove_after_bang(line):
    match = re_comment.match(line)
    if match is None:
      return line
    else:
      return re_comment.split(line)[1].rstrip()

  if form == Free_form:
    for line in text:
      if type(line) in [ Openmp, Doc, Directive] :
         result.append(line)
      elif type(line) == Empty_line:
         pass
      else:
        newline = line.text.lstrip()
        if newline == "" or newline[0] == "!":
          pass
        else:
          line.text = remove_after_bang(line.text)
          result.append(line)
    return result
  else:
    for line in text:
      if type(line) in [ Openmp, Doc, Directive ]:
         result.append(line)
      elif type(line) == Empty_line:
         pass
      else:
        newline = line.text.lstrip()
        if newline == "" or newline[0] == "!":
          pass
        else:
          line.text = remove_after_bang(line.text)
          if line.text[0] in "#123456789 ":
            result.append(line)
    return result

######################################################################
def remove_continuation(text,form):
  '''Removes continuation lines'''
  result = []
  buffer = ""
  number = 0
  t = None
  if form == Free_form:
    for line in text:
      if line.text[-1] == '&':
        buffer = "%s%s\n"%(buffer,line.text)
        if number == 0:
          t = type(line)
          number = line.i
      else:
        if number != 0:
          newline = t(number, \
            "%s%s"%(buffer,line.text), \
            line.filename)
          line = newline
          number = 0
          buffer = ""
        result.append(line)
  else:
    rev_text = list(text)
    rev_text.reverse()
    for line in rev_text:
      is_continuation = False
      if type(line) == Simple_line:
        if len(line.text) >= 6:
          if line.text[5] != ' ':
            is_continuation = True
      if is_continuation:
         buffer = "&\n%s %s %s"%(line.text[:5],line.text[6:],buffer)
      else:
         line.text = line.text+buffer
         result.insert(0,line)
         buffer = ""
  return result


######################################################################
def irp_simple_statements(text):
  '''Processes simple statements'''

  def process_irp_rw(line,rw,t):
    assert type(line) == t
    buffer = line.text.split()
    if len(buffer) == 2:
      dummy, variable = buffer
      num = "0"
    elif len(buffer) == 3:
      dummy, variable, num = buffer
    else:
      error.fail(line,"Error in IRP_%s statement"%(rw,))
    variable = variable.lower()
    i = line.i
    f = line.filename
    txt = line.text.lstrip()
    result = [
       Empty_line(i,"!",f),
       t(i,"! >>> %s"%(txt,),variable ),
       Provide_all(i,"   call %ser_%s('%s')"%(rw,variable,num),f),
       Empty_line(i,"! >>> END %s "%(txt,),f ),
       Empty_line(line.i,"!",f),
    ]
    return result

  def process_irp_read (line):
    assert type(line) == Irp_read
    return process_irp_rw(line,'read' ,Irp_read )

  def process_irp_write(line):
    assert type(line) == Irp_write
    return process_irp_rw(line,'writ' ,Irp_write)

  def process_return(line):
    assert type(line) == Return
    if command_line.do_debug:
      newline = Simple_line(line.i," call irp_leave(irp_here)",line.filename)
      result = [newline, line]
    else:
      result = [ line ]
    return result

  def debug_conditions(line):
    '''Find condition in assert statement for debug'''
    assert type(line) == Assert
    match = re_test.search(line.text)
    result = []
    if match is not None:
      matches = [ match.group(1).strip(), match.group(3).strip() ]
      for m in matches:
        ok = m != ""                              # not empty
        ok = ok and not m.isdigit()               # not a digit
        ok = ok and "'" not in m                  # not a string
        ok = ok and m.count('(') == m.count(')')  # balanced parenthesis
        if ok:
          result.append ( Simple_line (line.i, " print *, '%s = ', %s"%(m,m), line.filename) )
      result.append ( Simple_line (line.i, " print *, ''", line.filename) )
    return result

  def process_assert(line):
    assert type(line) == Assert
    if command_line.do_assert:
      if '(' not in line.text or ')' not in line.text:
         error.fail(line,"Syntax error in ASSERT statement (parentheses)")
      condition = "(%s"%(line.text.split('(',1)[1])
      if condition == "":
        error.fail(line,"Error in Assert statement")
      condition_str = condition.replace("'","''")
      i = line.i
      f = line.filename
      txt = line.text.strip()
      result = [
       Empty_line(i, "!", f),
       Empty_line(i, "! >>> %s"%(txt,), f),
       If         (i, "  if (.not.%s) then"%(condition,), f),
       Simple_line(i, "   print *, irp_here//': Assert failed:'", f),
       Simple_line(i, "   print *, ' file: %s, line: %d'"%(f,i), f),
       Simple_line(i, "   print *, '%s'"%(condition_str,), f),
       ] + debug_conditions(line) + [
       Simple_line(i, "   stop 1", f),
       Endif      (i, "  endif", f),
       Empty_line(i, "! <<< END %s"%(txt,), f),
       Empty_line(i, "!", f)
      ]
    else:
      result = []
    return result

  def process_end(line):
    '''Set irp_here variable in provider block'''
    line.text = "end"
    if command_line.do_debug:
      i = line.i
      f = line.filename
      result = [
          Simple_line(i," call irp_leave(irp_here)", f),
          line
        ]
    else:
      result = [ line ]
    return result

  def process_begin_provider(line):
    assert type(line) == Begin_provider
    buffer = line.lower.replace('['," ")
    buffer = buffer.replace(']',"")
    buffer = buffer.split(',')
    if len(buffer) < 2:
      error.fail(line,"Error in Begin_provider statement")
    varname = buffer[1].strip()
    length = len(varname)
    i = line.i
    f = line.filename
    result = [ Begin_provider(i,line.text, (f,varname)),
      Declaration(i,"  character*(%d) :: irp_here = '%s'"%(length,varname), f) ]
    if command_line.do_debug:
      result += [
        Simple_line(i,"  call irp_enter(irp_here)", f),
      ]
    return result

  def process_cont_provider(line):
    assert type(line) == Cont_provider
    buffer = line.lower.replace('['," ")
    buffer = buffer.replace(']',"")
    buffer = buffer.split(',')
    if len(buffer) < 2:
      error.fail(line,"Error in Cont_provider statement")
    varname = buffer[1].strip()
    i = line.i
    f = line.filename
    return [ Cont_provider(i,line.text,(f,varname)) ]

  def process_subroutine(line):
    assert type(line) == Subroutine
    subname = find_subname(line)
    length = len(subname)
    i = line.i
    f = line.filename
    result = [ line,
      Declaration(i,"  character*(%d) :: irp_here = '%s'"%(length,subname), f) ]
    if command_line.do_debug:
      result += [
        Simple_line(i,"  call irp_enter_f(irp_here)", f),
      ]
    return result

  def process_function(line):
    assert type(line) == Function
    subname = find_subname(line)
    length = len(subname)
    i = line.i
    f = line.filename
    result = [ line,
      Declaration(i,"  character*(%d) :: irp_here = '%s'"%(length,subname), f) ]
    if command_line.do_debug:
      result += [
        Simple_line(i,"  call irp_enter_f(irp_here)", f),
      ]
    return result


  def process_program(line):
    assert type(line) == Program
    program_name = line.lower.split()[1]
    temp = [ Program(0,"program irp_program",program_name) ]
    if command_line.do_profile:
      temp += [ Simple_line(0,"call irp_init_timer()",line.filename) ]
    if command_line.do_openmp:
      temp += [ Simple_line(0," call irp_init_locks_%s()"%(irp_id),line.filename) ]
    temp += [ Call(0," call %s"%(program_name),line.filename) ]
    if command_line.do_profile:
      temp += [ Simple_line(0,"call irp_print_timer()",line.filename) ]
    temp += [ Simple_line(0," call irp_finalize_%s()"%(irp_id),line.filename) ]
    temp += [ End(0,"end program",line.filename) ]
    result = temp + \
       process_subroutine( Subroutine(line.i,"subroutine %s"%(program_name,),line.filename) )
    return result

  d = { Irp_read       : process_irp_read,
        Irp_write      : process_irp_write,
        Return         : process_return,
        Assert         : process_assert,
        End            : process_end,
        Begin_provider : process_begin_provider,
        Cont_provider  : process_cont_provider,
        End_provider   : process_end,
        Subroutine     : process_subroutine,
        Function       : process_function,
        Program        : process_program,
      }

  result = []
  for line in text:
    buffer = [ line ]
    for t in d:
      if type(line) == t:
        buffer = d[t](line)
        break
    result += buffer
  return result


######################################################################
def change_includes(text):
  '''Deals with include files'''
  result = []
  for line in text:
    if type(line) == Include:
     txt = line.text.replace('"',"'").split("'")
     if len(txt) != 3:
       print txt
       error.fail(line,"Error in include statement")
     directory = (("./"+line.filename).rsplit('/',1)[0]+'/')[2:]
     if directory == "":
        filename = txt[1].strip()
     else:
        filename = directory+txt[1].strip()
     try:
       file = open(filename,'r')
       file.close()
       result.append(Include(line.i,"! include '%s'"%filename,filename))
       result += create_preprocessed_text(filename)
     except IOError:
       result.append(Declaration(line.i,line.text,line.filename))
    else:
     result.append(line)
  return result

######################################################################
def process_old_style_do(text):
  '''Changes old-style do loops to new style'''
  assert type(text) == list

  def change_matching_enddo(begin,number):
    for i in range(begin+1,len(text)):
      line = text[i]
      if type(line) in [Continue,Enddo]:
        buffer = line.text.split()
        if buffer[0] == number:
          text[i] = Enddo(line.i,"  enddo",line.filename)
          return
    error.fail(text[begin],"Old-style do loops should end with 'continue' or 'end do'")

  result = []
  for i in range(len(text)):
    line = text[i]
    if type(line) == Do:
      buffer = line.text.split()
      try:
       if buffer[1].isdigit():
        number = buffer.pop(1)
        change_matching_enddo(i,number)
        line.text = " ".join(buffer)
      except IndexError:
        pass
    result.append(line)
  return result

######################################################################
def change_single_line_ifs(text):
  '''Changes:
if (test) result

to

if (test) then
  result
endif'''

  assert type(text) == list
  result = []
  for line in text:
    if type(line) == If:
      if line.lower.endswith("then"):
        result.append(line)
      else:
        buffer = line.text
        begin = buffer.find('(')
        if begin < 0:
          error.fail(line,"Error in if statement")
        level = 0
        instring = False
        for i,c in enumerate(buffer[begin:]):
          if c == "'":
            instring = not instring
          if instring:
            pass
          elif c == '(':
            level +=1
          elif c == ')':
            level -= 1
          if level == 0:
            end = begin+i+1
            break
        if level != 0:
          error.fail(line,"Error in if statement")
        test = buffer[:end]
        code = buffer[end:]
        i = line.i
        f = line.filename
        result.append( If(i,"%s then"%(test,),f) )
        result += get_type(i,f,code,False)[0]
        result.append( Endif(i,"  endif",f) )
    else:
      result.append(line)
  return result

######################################################################
def check_begin_end(text):
  '''Checks x...endx consistence'''

  filter_line = lambda line: type(line) in [ Do, Enddo, If, Endif, \
         Program, Begin_provider, End_provider, \
         Subroutine, Function, End, Begin_doc, End_doc ]
  text = filter(filter_line, text)

  d = { 'do' : Do,         'enddo': Enddo,
        'if' : If,         'endif': Endif,
        '_doc': Begin_doc, 'end_doc': End_doc}
  assert type(text) == list

  def find_matching_end_ifdo(begin,x):
    level = 1
    for i in range(begin+1,len(text)):
      line = text[i]
      if type(line) == d[x]:
        level += 1
      elif type(line) == d["end%s"%(x,)]:
        level -= 1
        if level == 0:
          return True
      elif type(line) in [End, End_provider]:
        break
    error.fail(text[begin],"Missing 'end%s'"%(x,))

  def find_matching_end_subfunpro(begin,x):
    line = text[begin]
    for i in range(begin+1,len(text)):
      line = text[i]
      if type(line) == x:
        return
      if type(line) in [ Subroutine, Function, Program, Begin_provider ]:
        error.fail(text[begin],type(line).str+" is not closed")
    error.fail(text[begin],type(line).str + " is not closed")


  level = 0
  for i,line in enumerate(text):
    if type(line) == Begin_doc:
      find_matching_end_ifdo(i,'_doc')
  for i,line in enumerate(text):
    if type(line) == Do:
      find_matching_end_ifdo(i,'do')
    elif type(line) == If:
      find_matching_end_ifdo(i,'if')
    elif type(line) in [Subroutine, Function, Program]:
      level += 1
      find_matching_end_subfunpro(i,End)
    elif type(line) == Begin_provider:
      level += 1
      find_matching_end_subfunpro(i,End_provider)
    elif type(line) == End:
      level -= 1
    elif type(line) == End_provider:
      level -= 1
    if level < 0:
      error.fail(line,"Beginning of block not matched")

  return True

######################################################################
def remove_ifdefs(text):
  assert type(text) == list
  result = []
  do_print = True
  for line in text:
    if type(line) == Irp_If:
      var = line.text.split()[1]
      do_print = var in command_line.defined
    elif type(line) == Irp_Else:
      do_print = not do_print
    elif type(line) == Irp_Endif:
      do_print = True
    else:
      if do_print:
        result.append(line)
  return result

######################################################################
def check_OpenMP(text):
  assert type(text) == list
  result = []
  inside_openmp = False
  for line in text:
     if type(line) == Openmp:
       # Detect OpenMP blocks
       buffer = line.text.lower().split()
       if buffer[1] == "parallel":
         inside_openmp = True
       if buffer[1] == "end" and buffer[2] == "parallel":
         inside_openmp = False
     result.append(line)

     if inside_openmp:
       if type(line) in [ Provide_all, Provide, Touch, SoftTouch ]:
          error.fail(line,type(line).str+" is not allowed in an OpenMP block.")

  return result

######################################################################
def create_preprocessed_text(filename):
  file = open(filename,"r")
  lines = file.readlines()
  file.close()
  result = get_text(lines,filename)
  result = execute_templates(result)
  result = execute_shell(result)
  fortran_form = form(result)
  result = remove_ifdefs(result)
  result = remove_comments(result,fortran_form)
  result = remove_continuation(result,fortran_form)
  result = add_operators(result)
  result = change_includes(result)
  result = change_single_line_ifs(result)
  result = process_old_style_do(result)
  result = irp_simple_statements(result)
  result = check_OpenMP(result)
  check_begin_end(result)
  return result

######################################################################
preprocessed_text = parallel_loop( lambda x,y: create_preprocessed_text(x), \
                 map(lambda x: (x,None), irpf90_files ) )

######################################################################
def debug():
  for filename, txt in preprocessed_text:
   if filename == 'invert.irp.f':
    print "=== "+filename+" ==="
    for line in txt:
      print line
  print irpf90_files

if __name__ == '__main__':
  debug()




#!/usr/bin/env python2
#   IRPF90 is a Fortran90 preprocessor written in Python for programming using
#   the Implicit Reference to Parameters (IRP) method.
#   Copyright (C) 2009 Anthony SCEMAMA 
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License along
#   with this program; if not, write to the Free Software Foundation, Inc.,
#   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
#   Anthony Scemama
#   LCPQ - IRSAMC - CNRS
#   Universite Paul Sabatier
#   118, route de Narbonne      
#   31062 Toulouse Cedex 4      
#   scemama@irsamc.ups-tlse.fr


rdtsc = """
#ifdef __i386
double irp_rdtsc_(void) {
  unsigned long long x;
  __asm__ volatile ("rdtsc" : "=A" (x));
  return (double) x;
}
#elif __amd64
double irp_rdtsc_(void) {
  unsigned long long a, d;
  __asm__ volatile ("rdtsc" : "=a" (a), "=d" (d));
  return (double)((d<<32) | a);
}
#endif
"""

import subprocess
import tempfile
import os
import threading
from irpf90_t import irpdir

def build_rdtsc():
  filename = irpdir+"irp_rdtsc.c"
  file = open(filename,'w')
  file.write(rdtsc)
  file.close()
  def t():
    p = subprocess.Popen(["gcc","-O2",filename,"-c","-o","irp_rdtsc.o"])
    p.communicate()
    os.remove(filename)

  threading.Thread(target=t).start()

def build_module():
  from variables import variables
  data = """
module irp_timer
 double precision :: irp_profile(3,%(n)d) 
 integer          :: irp_order(%(n)d) 
 character*(64)   :: irp_profile_label(%(n)d)
 double precision :: irp_rdtsc_shift

 contains

   subroutine profile_sort ()
    implicit none
    character*(64)         :: xtmp
    integer                :: i, i0, j, jmax

    do i=1,size(irp_profile_label)
     irp_order(i)=i
    enddo
    do i=1,size(irp_profile_label)
     xtmp = irp_profile_label(i)
     i0 = irp_order(i)
     j = i-1
     do j=i-1,1,-1
      if ( irp_profile_label(j) > xtmp ) then
       irp_profile_label(j+1) = irp_profile_label(j)
       irp_order(j+1) = irp_order(j)
      else
       exit
      endif
     enddo
     irp_profile_label(j+1) = xtmp
     irp_order(j+1) = i0
    enddo
   end subroutine profile_sort

end module

subroutine irp_init_timer
 use irp_timer
 implicit none
 integer :: i
 double precision :: irp_rdtsc, t0
 irp_profile = 0.d0
 irp_rdtsc_shift = 0.d0
 do i=1,1000000
   t0 = irp_rdtsc()
 enddo
 irp_rdtsc_shift = 1.d-6*(irp_rdtsc()-t0)
%(text)s
end

subroutine irp_set_timer(i,value)
 use irp_timer
 implicit none
 integer, intent(in) :: i
 double precision, intent(inout) :: value
 value = value - irp_rdtsc_shift
 irp_profile(1,i) = irp_profile(1,i) + value
 irp_profile(2,i) = irp_profile(2,i) + value*value
 irp_profile(3,i) = irp_profile(3,i) + 1.d0
end

subroutine irp_print_timer()
 use irp_timer
 implicit none
 integer :: i, ii
 double precision :: error, sigma2, average, average2, frequency, t0
 double precision :: irp_rdtsc
 t0 = irp_rdtsc()
 call sleep(1)
 frequency = (irp_rdtsc()-t0-irp_rdtsc_shift)

 call profile_sort()
 print '(A24,A8,A17,A20,A13,A20)', '', 'N.Calls', 'Tot Cycles', 'Avg Cycles', &
                          'Tot Secs', 'Avg Secs'
   print '(A)', '---------------------------------------------------'// &
                '---------------------------------------------------'
 do ii=1,%(n)d
  i = irp_order(ii)
  if (irp_profile(3,i) > 0.) then
   error = 0.d0
   average  = irp_profile(1,i)/irp_profile(3,i)
   if (irp_profile(3,i) > 1.d0) then
     average2 = irp_profile(2,i)/irp_profile(3,i)
     sigma2   = (average2 - average*average)
     error = sqrt(sigma2/(irp_profile(3,i)+1.d0))
   endif
   print '(A24 , F8.0 , X,F12.0 , X,F12.0,A3,F8.0, X,F12.8, X,F8.5,A3,F8.5 )', &
     irp_profile_label(ii), &
     irp_profile(3,i), &
     irp_profile(1,i), &
     average, '+/-', error, &
     irp_profile(1,i)/frequency, &
     average/frequency, '+/-', error/frequency
  endif
 enddo
 print *, 'Frequency     :', frequency*1.d-9, ' GHz'
 print *, 'rdtsc latency :', irp_rdtsc_shift, ' cycles'
end
  """
  label = {}
  for i in variables:
    vi = variables[i]
    label[vi.label] = vi.same_as
  text = []
  lmax = 0
  for l in label:
    text.append(" irp_profile_label(%d) = '%s'"%(l,label[l]))
    lmax = max(lmax,l)
  text.sort()
  text = '\n'.join(text)
  data = data%{'text': text, 'n':lmax}
  file = open("IRPF90_temp/irp_profile.irp.F90",'w')
  file.write(data)
  file.close()

def run():
  build_module()
  build_rdtsc()

if __name__ == "__main__":
  build_rdtsc()



#!/usr/bin/env python2
#   IRPF90 is a Fortran90 preprocessor written in Python for programming using
#   the Implicit Reference to Parameters (IRP) method.
#   Copyright (C) 2009 Anthony SCEMAMA 
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License along
#   with this program; if not, write to the Free Software Foundation, Inc.,
#   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
#   Anthony Scemama
#   LCPQ - IRSAMC - CNRS
#   Universite Paul Sabatier
#   118, route de Narbonne      
#   31062 Toulouse Cedex 4      
#   scemama@irsamc.ups-tlse.fr


from irpf90_t import *
from variable import *
from variables import variables
from command_line import command_line
import preprocessed_text
from util import *

class Fmodule(object):

  header = \
     [ "! -*- F90 -*-",
       "!",
       "!-----------------------------------------------!",
       "! This file was generated with the irpf90 tool. !",
       "!                                               !",
       "!           DO NOT MODIFY IT BY HAND            !",
       "!-----------------------------------------------!",
       "" ]

  def __init__(self,text,filename):
    self.text = put_info(text,filename)
    self.filename = filename[:-6]
    self.name = "%s_mod"%(self.filename).replace('/','__').replace('.','Dot')

  def is_main(self):
    if '_is_main' not in self.__dict__:
      self._is_main = self.prog_name is not None
    return self._is_main
  is_main = property(is_main)

  def prog_name(self):
    if '_prog_name' not in self.__dict__:
      buffer = filter(lambda x: type(x[1]) == Program,self.text)
      if buffer == []:
        self._prog_name = None
      else:
        self._prog_name = buffer[0][1].filename
    return self._prog_name
  prog_name = property(prog_name)

  def variables(self):
    if '_variables' not in self.__dict__:
      from variables import variables
      name = self.name
      self._variables = filter(lambda x: variables[x].fmodule == name, variables)
    return self._variables
  variables = property(variables)


  def head(self):
    if '_head' not in self.__dict__:
      result = [ "module %s"%(self.name) ]
      result += self.use
      result += self.dec
      result += flatten( map(lambda x: variables[x].header,self.variables) )
      result.append( "end module %s"%(self.name) )
      self._head = result
    return self._head
  head = property(head)

  def needed_vars(self):
    if '_needed_vars' not in self.__dict__:
      result = map(lambda x: variables[x].needs,self.variables)
      result = make_single ( flatten(result) )
      self._needed_vars = result
    return self._needed_vars
  needed_vars = property(needed_vars)

  def includes(self):
    if '_includes' not in self.__dict__:
      buffer = []
      for v in self.needed_vars:
        buffer += variables[v].includes
      self._includes = make_single(buffer)
    return self._includes
  includes = property(includes)

  def generated_text(self):
    if '_generated_text' not in self.__dict__:
      result = []
      for var in self.variables:
        var = variables[var]
        result += var.provider
        result += var.builder
        if var.is_read:
          result += var.reader
        if var.is_written:
          result += var.writer
      self._generated_text = result
    return self._generated_text
  generated_text = property(generated_text)

  def residual_text(self):
    if '_residual_text' not in self.__dict__:
      from variables import build_use, call_provides
      from parsed_text import move_to_top
      def remove_providers(text):
        result = []
        inside = False
        for vars,line in text:
          if type(line) == Begin_provider:
            inside = True
          if not inside:
            result.append( (vars,line) )
          if type(line) == End_provider:
            inside = False
        return result

      def modify_functions(text):
        result = []
        variable_list = []
        for vars,line in text:
          if type(line) in [ Subroutine, Function ]:
            variable_list = list(vars)
          elif type(line) == End:
            result += map(lambda x: ([],Use(line.i,x,line.filename)), build_use(variable_list))
          else:
            variable_list += vars
          result.append( (vars,line) )
        return result

      def extract_use_dec_text(text):
        inside = False
        result = []
        dec = []
        use = []
        for vars,line in text:
          if type(line) in [ Subroutine, Function, Program]:
            inside = True
          if inside:
            result.append( (vars,line) )
          else:
            if type(line) == Use:
              use.append( (vars,line) )
            elif type(line) == Declaration:
              dec.append( (vars,line) )
          if type(line) == End:
            inside = False
        return use, dec, result

      def provide_variables(text):
        result = []
        for vars,line in text:
          result.append( ([],line) )
          result += map(lambda x: ([],Simple_line(line.i,x,line.filename)), call_provides(vars)) 
        return result

      result = remove_providers(self.text)
      result = modify_functions(result)
      use,dec,result = extract_use_dec_text(result)
      self._use = make_single(map(lambda x: " "+x[1].text, use))
      self._dec = map(lambda x: " "+x[1].text, dec)
#      self._dec = make_single(map(lambda x: " "+x[1].text, dec))
      result = provide_variables(result)
      result = move_to_top(result,Declaration)
      result = move_to_top(result,Implicit)
      result = move_to_top(result,Use)
      result    = map(lambda x: x[1], result)
      result    = map(lambda x: x.text, result)
      self._residual_text = result
    return self._residual_text
  residual_text = property(residual_text)

  def use(self):
    if '_use' not in self.__dict__:
      self.residual_text
    return self._use
  use = property(use)

  def dec(self):
    if '_dec' not in self.__dict__:
      self.residual_text
    return self._dec
  dec = property(dec)


  def needed_modules(self):
    if '_needed_modules' not in self.__dict__:
      buffer = filter(lambda x: x.lstrip().startswith("use "), \
        self.generated_text+self.head+self.residual_text)
      buffer = map(lambda x: x.split()[1], buffer)
      buffer = filter(lambda x: x.endswith("_mod"),buffer )
      self._needed_modules = make_single(buffer)
      if self.name in self._needed_modules:
        self._needed_modules.remove(self.name)
    return self._needed_modules
  needed_modules = property(needed_modules)

######################################################################

if __name__ == '__main__':
  from parsed_text import parsed_text
  for filename, text in parsed_text:
    if filename == 'vmc_step.irp.f':
     x = Fmodule(text,filename)
     break
  for line in x.head:
    print line
  print x.includes




#!/usr/bin/env python2
#   IRPF90 is a Fortran90 preprocessor written in Python for programming using
#   the Implicit Reference to Parameters (IRP) method.
#   Copyright (C) 2009 Anthony SCEMAMA 
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License along
#   with this program; if not, write to the Free Software Foundation, Inc.,
#   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
#   Anthony Scemama
#   LCPQ - IRSAMC - CNRS
#   Universite Paul Sabatier
#   118, route de Narbonne      
#   31062 Toulouse Cedex 4      
#   scemama@irsamc.ups-tlse.fr


from irpf90_t import *
from util import *
import error
from command_line import command_line

class Variable(object):

  ############################################################
  def __init__(self,text,label,name = None):
    assert type(text) == list
    assert len(text) > 0
    assert type(text[0]) == Begin_provider
    self.label = label
    self.text = text
    if name is not None:
      self._name = name

  ############################################################
  def is_touched(self):
    if '_is_touched' not in self.__dict__:
      from variables import variables
      result = self.is_read
      for i in self.children:
        if variables[i].is_touched:
          result = True
          break
      self._is_touched = result
    return self._is_touched
  is_touched = property(is_touched)

  ############################################################
  def has_openmp(self):
    if '_has_openmp' not in self.__dict__:
      self._has_openmp = False
      self.builder()
    return self._has_openmp
  has_openmp = property(has_openmp)

  ############################################################
  def is_written(self):
    if '_is_written' not in self.__dict__:
      from variables import variables
      result = False
      for i in self.parents:
        if variables[i].is_written:
          result = True
          break
      self._is_written = result
    return self._is_written
  is_written = property(is_written)

  ############################################################
  def is_read(self):
    if '_is_read' not in self.__dict__:
      from variables import variables
      result = False
      for i in self.parents:
        if variables[i].is_read:
          result = True
          break
      self._is_read = result
    return self._is_read
  is_read = property(is_read)

  ############################################################
  def is_main(self):
    if '_is_main' not in self.__dict__:
      self._is_main = (self.name == self.same_as)
    return self._is_main
  is_main = property(is_main)

  ############################################################
  def name(self):
    '''Name is lowercase'''
    if '_name' not in self.__dict__:
      buffer = None
      text = self.text
      for line in text:
        if type(line) == Begin_provider:
          self._name = line.filename[1]
          if "(" in self._name:
            error.fail(line,"""'(' not allowed in variable name.
A ',' is probably missing between the name of the variable and the
dimension.
""") 
          break
    return self._name
  name = property(name)

  ############################################################
  def doc(self):
    if '_doc' not in self.__dict__:
      text = self.text
      buffer = filter(lambda l:type(l) == Doc, text)
      self._doc = map(lambda l: l.text.lstrip()[1:], buffer)
      if buffer == []:
        error.warn(None,"Variable %s is not documented"%(self.name))
    return self._doc
  doc = property(doc)

  ############################################################
  def documented(self):
    if '_documented' not in self.__dict__:
      self._documented = (self.doc != [])
    return self._documented
  documented = property(documented)

  ############################################################
  def includes(self):
    if '_includes' not in self.__dict__:
      self._includes = []
      text = self.text
      for line in filter(lambda x: type(x) == Include,text):
        self._includes.append(line.filename)
      make_single(self._includes)
    return self._includes
  includes = property(includes)

  ############################################################
  def calls(self):
    if '_calls' not in self.__dict__:
      self._calls = []
      text = self.text
      for line in filter(lambda x: type(x) == Call,text):
        sub = line.text.split('(',1)[0].split()[1].lower()
        self._calls.append(sub)
      make_single(self._calls)
    return self._calls
  calls=property(fget=calls)

  ############################################################
  def others(self):
    if '_others' not in self.__dict__:
      result = []
      append = result.append
      f = lambda  l: type(l) in [Begin_provider, Cont_provider]
      text = self.text
      lines = filter(f, text)
      for line in lines:
        append(line.filename[1])
      result.remove(self.name)
      self._others = result
    return self._others
  others = property(others)

  ############################################################
  def same_as(self):
    if '_same_as' not in self.__dict__:
      if type(self.line) == Begin_provider:
        result = self.name
      else:
        result = self.text[0].filename[1]
      self._same_as = result
    return self._same_as
  same_as = property(same_as)

  ############################################################
  def allocate(self):
    if '_allocate' not in self.__dict__:
      if not self.is_main:
        self._allocate = []
      else:
        from variables import variables
        def f(var):
          return variables[var].dim != []
        self._allocate = filter ( f, self.others + [self.name] )
    return self._allocate
  allocate = property(allocate)

  ############################################################
  def dim(self):
    if '_dim' not in self.__dict__:
      line = self.line.text.split('!')[0]
      buffer = line.replace(']','').split(',',2)
      if len(buffer) == 2:
        self._dim = []
      else:
        buffer = buffer[2].strip()[1:-1].split(',')
        self._dim = map(strip,buffer)
    return self._dim
  dim = property(dim)

  ############################################################
  def type(self):
    if '_type' not in self.__dict__:
      line = self.line.text
      buffer = line.split(',')[0]
      try:
        buffer = buffer.split('[')[1].strip()
      except IndexError:
        error.fail(None,"Error in definition of %s."%(self.name))
      if self.dim != []:
        buffer = "%s, allocatable"%(buffer)
      self._type = buffer
    return self._type
  type = property(type)

  ############################################################
  def fmodule(self):
    if '_fmodule' not in self.__dict__:
      self._fmodule = self.line.filename[0].replace('/','__').split('.irp.f')[0]+'_mod'
    return self._fmodule
  fmodule = property(fmodule)

  ############################################################
  def regexp(self):
    if '_regexp' not in self.__dict__:
      import re
      self._regexp = re.compile( \
        r"([^a-z0-9'\"_]|^)%s([^a-z0-9_]|$)"%(self.name),re.I)
    return self._regexp
  regexp = property(regexp)

  ############################################################
  def line(self):
    if '_line' not in self.__dict__:
      f = lambda l: type(l) in [Begin_provider, Cont_provider]
      text = self.text
      lines = filter(f, text)
      for line in lines:
        buffer = line.filename[1]
        if self._name == buffer:
          self._line = line
          break
    assert '_line' in self.__dict__
    return self._line
  line = property(line)

  ############################################################
  def header(self):
    if '_header' not in self.__dict__:
      name = self.name
      self._header = [ "  %s :: %s %s"%(self.type, name, build_dim_colons(self) ) ]
      if command_line.coarray:
        if self.dim == []:
          self._header[0] += " [*]"
        else:
          self._header[0] += " [:]"
      if self.dim != [] and command_line.align != '1':
       self._header += ["  !DIR$ ATTRIBUTES ALIGN: %s :: %s"%(command_line.align,name)]
      if self.is_main:
       self._header += [ "  logical :: %s_is_built = .False."%(name) ]
    return self._header
  header = property(header)

  ############################################################
  def toucher(self):
    if '_toucher' not in self.__dict__:
      if not self.is_main:
        self._toucher = []
      else:
        from modules import modules
        from variables import variables
        if '_needed_by' not in self.__dict__:
          import parsed_text
        parents = self.parents
        parents.sort()
        mods = map(lambda x: variables[x].fmodule, parents)
        mods = make_single(mods)+[self.fmodule]
        name = self.name
        result = [ "subroutine touch_%s"%(name) ]
        result += map(lambda x: "  Use %s"%(x),mods)
        result.append("  implicit none")
        if command_line.do_debug:
          length = str(len("touch_%s"%(name)))
          result += [ "  character*(%s) :: irp_here = 'touch_%s'"%(length,name) ]
        if command_line.do_debug:
          result += [ "  call irp_enter(irp_here)" ]
        result += map( lambda x: "  %s_is_built = .False."%(x), parents)
        result.append("  %s_is_built = .True."%(name))
        if command_line.do_debug:
          result.append("  call irp_leave(irp_here)")
        result.append("end subroutine touch_%s"%(name))
        result.append("")
        self._toucher = result
    return self._toucher
  toucher = property(toucher)

  ##########################################################
  def locker(self):
    if '_locker' not in self.__dict__:
      if not command_line.do_openmp:
        self._locker = []
      else:
        from modules import modules
        from variables import variables
        name = self.name
        result  = [ "subroutine irp_lock_%s(set)"%(name) ]
        result += [ "  use omp_lib",
                    "  implicit none",
                    "  logical, intent(in) :: set",
                    "  integer(kind=omp_nest_lock_kind),save :: %s_lock"%(name), 
                    "  integer,save :: ifirst",
                  ]
        if command_line.do_debug:
          length = str(len("irp_lock_%s"%(name)))
          result += [ "  character*(%s) :: irp_here = 'irp_lock_%s'"%(length,name),
                      "  call irp_enter(irp_here)" ]
        result += [ "  if (ifirst == 0) then",
                    "    ifirst = 1",
                    "    call omp_init_nest_lock(%s_lock)"%(name),
                    "  endif",
                    "  if (set) then",
                    "    call omp_set_nest_lock(%s_lock)"%(name),
                    "  else",
                    "    call omp_unset_nest_lock(%s_lock)"%(name),
                    "  endif",
                  ]
        if command_line.do_debug:
          result.append("  call irp_leave(irp_here)")
        result.append("end subroutine irp_lock_%s"%(name))
        result.append("")
        self._locker = result
    return self._locker
  locker = property(locker)

  ##########################################################
  def reader(self):
    if '_reader' not in self.__dict__:
      if not self.is_main:
        self._reader = []
      else:
        if '_needs' not in self.__dict__:
          import parsed_text
        from variables import variables
        name = self.name
        result = [ \
        "subroutine reader_%s(irp_num)"%(name),
        "  use %s"%(self.fmodule),
        "  implicit none",
        "  character*(*), intent(in) :: irp_num",
        "  logical                   :: irp_is_open",
        "  integer                   :: irp_iunit" ]
        if command_line.do_debug:
          length = len("reader_%s"%(self.name))
          result += [\
          "  character*(%d) :: irp_here = 'reader_%s'"%(length,name),
          "  call irp_enter(irp_here)" ]
        result += map(lambda x: "  call reader_%s(irp_num)"%(x),self.needs) 
        result += [ \
        "  irp_is_open = .True.",
        "  irp_iunit = 9",
        "  do while (irp_is_open)",
        "   irp_iunit = irp_iunit+1", 
        "   inquire(unit=irp_iunit,opened=irp_is_open)",
        "  enddo"]
        for n in [ name ]+self.others:
          result += [\
          "  open(unit=irp_iunit,file='irpf90_%s_'//trim(irp_num),form='FORMATTED',status='OLD',action='READ')"%(n),
          "  read(irp_iunit,*) %s%s"%(n,build_dim_colons(variables[n])),
          "  close(irp_iunit)" ]
        result += [ \
        "  call touch_%s"%(name),
        "  %s_is_built = .True."%(name) ]
        if command_line.do_debug:
          result.append("  call irp_leave(irp_here)")
        result.append("end subroutine reader_%s"%(name))
        result.append("")
        self._reader = result
    return self._reader
  reader = property(reader)

  ##########################################################
  def writer(self):
    if '_writer' not in self.__dict__:
      if not self.is_main:
        self._writer = []
      else:
        from variables import variables
        if '_needs' not in self.__dict__:
          import parsed_text
        name = self.name
        result = [ \
        "subroutine writer_%s(irp_num)"%(name),
        "  use %s"%(self.fmodule),
        "  implicit none",
        "  character*(*), intent(in) :: irp_num",
        "  logical                   :: irp_is_open",
        "  integer                   :: irp_iunit" ]
        if command_line.do_debug:
          length = len("writer_%s"%(self.name))
          result += [\
          "  character*(%d) :: irp_here = 'writer_%s'"%(length,name),
          "  call irp_enter(irp_here)" ]
        result += [ \
        "  if (.not.%s_is_built) then"%(self.same_as),
        "    call provide_%s"%(self.same_as),
        "  endif" ]
        result += map(lambda x: "  call writer_%s(irp_num)"%(x),self.needs) 
        result += [ \
        "  irp_is_open = .True.",
        "  irp_iunit = 9",
        "  do while (irp_is_open)",
        "   irp_iunit = irp_iunit+1", 
        "   inquire(unit=irp_iunit,opened=irp_is_open)",
        "  enddo" ]
        for n in [ name ] + self.others:
          result += [\
          "  open(unit=irp_iunit,file='irpf90_%s_'//trim(irp_num),form='FORMATTED',status='UNKNOWN',action='WRITE')"%(n),
          "  write(irp_iunit,*) %s%s"%(n,build_dim_colons(variables[n])),
          "  close(irp_iunit)" ]
        if command_line.do_debug:
          result.append("  call irp_leave(irp_here)")
        result.append("end subroutine writer_%s"%(name))
        result.append("")
        self._writer = result
    return self._writer
  writer = property(writer)

  ##########################################################
  def free(self):
    if '_free' not in self.__dict__:
      name = self.name
      result = [ "!","! >>> FREE %s"%(name),
        "  %s_is_built = .False."%(self.same_as) ] 
      if self.dim != []:
        if command_line.do_memory:
          result += [ \
          "  if (allocated(%s)) then"%(name),
          "    deallocate (%s)"%(name),
          "    print *, 'Deallocating %s'"%(name),
          "  endif" ]
        else:
          result += [ \
          "  if (allocated(%s)) then"%(name),
          "    deallocate (%s)"%(name),
          "  endif" ]
      result.append("! <<< END FREE")
      self._free = result
    return self._free
  free = property(free)

  ##########################################################
  def provider(self):
    if '_provider' not in self.__dict__:
     if not self.is_main:
       self._provider = []
     else:
      if '_to_provide' not in self.__dict__:
        import parsed_text
      from variables import variables, build_use, call_provides
      name = self.name
      same_as = self.same_as

      def build_alloc(name):
        self = variables[name]
        if self.dim == []:
           return []

        def do_size():
           result = "     print *, ' size: ("
           result += ','.join(self.dim)
           return result+")'"

        def check_dimensions():
          result = map(lambda x: "(%s>0)"%(dimsize(x)), self.dim)
          result = ".and.".join(result)
          result = "   if (%s) then"%(result)
          return result
 
        def dimensions_OK():
          result = [ "  irp_dimensions_OK = .True." ]
          for i,k in enumerate(self.dim):
              result.append("  irp_dimensions_OK = irp_dimensions_OK.AND.(SIZE(%s,%d)==(%s))"%(name,i+1,dimsize(k)))
          return result

        def do_allocate():
          if command_line.coarray:
            result = "    allocate(%s(%s)[*],stat=irp_err)"
          else:
            result = "    allocate(%s(%s),stat=irp_err)"
          result = result%(name,','.join(self.dim))
          if command_line.do_memory:
            tmp = "\n   print *, %s, 'Allocating %s(%s)'"
            d = ','.join(self.dim)
            result += tmp%('size('+name+')',name,d)
          return result

        result = [ " if (allocated (%s) ) then"%(name) ]
        result += dimensions_OK()
        result += [\
          "  if (.not.irp_dimensions_OK) then",
          "   deallocate(%s,stat=irp_err)"%(name),
          "   if (irp_err /= 0) then",
          "     print *, irp_here//': Deallocation failed: %s'"%(name),
          do_size(),
          "   endif"]
        if command_line.do_memory:
          result += [\
          "   print *, 'Deallocating %s'"%(name) ]
        result.append(check_dimensions())
        result.append(do_allocate())
        result += [\
          "    if (irp_err /= 0) then",
          "     print *, irp_here//': Allocation failed: %s'"%(name),
          do_size(),
          "    endif",
          "   endif",
          "  endif",
          " else" ]
        result.append(check_dimensions())
        result.append(do_allocate())
        result += [\
          "    if (irp_err /= 0) then",
          "     print *, irp_here//': Allocation failed: %s'"%(name),
          do_size(),
          "    endif",
          "   endif",
          " endif" ]
        return result

      result = []
      if command_line.directives and command_line.inline in ["all","providers"]:
        result += [ "!DEC$ ATTRIBUTES FORCEINLINE :: provide_%s"%(name) ]
      result += [ "subroutine provide_%s"%(name) ] 
      result += build_use( [same_as]+self.to_provide )
      if command_line.do_openmp:
        result += [" use omp_lib"]
      result.append("  implicit none")
      length = len("provide_%s"%(name))
      result += [\
      "  character*(%d) :: irp_here = 'provide_%s'"%(length,name),
      "  integer                   :: irp_err ",
      "  logical                   :: irp_dimensions_OK",
      "!$ integer                  :: nthreads"]
      if command_line.do_openmp:
        result.append(" call irp_lock_%s(.True.)"%(same_as))
      if command_line.do_debug:
        result.append("  call irp_enter(irp_here)")
      result += call_provides(self.to_provide)
      result += flatten( map(build_alloc,[self.same_as]+self.others) )
      result += [ " if (.not.%s_is_built) then"%(same_as),
                  "  call bld_%s"%(same_as),
                  "  %s_is_built = .True."%(same_as), "" ]
      result += [ " endif" ]
      if command_line.do_debug:
        result.append("  call irp_leave(irp_here)")
      if command_line.do_openmp:
        result.append(" call irp_lock_%s(.False.)"%(same_as))
      result.append("end subroutine provide_%s"%(name) )
      result.append("")
      self._provider = result
    return self._provider
  provider = property(provider)

  ##########################################################
  def builder(self):
    if '_builder' not in self.__dict__:
      if not self.is_main:
        self._builder = []
      else:
        import parsed_text
        from variables import build_use, call_provides
        for filename,buffer in parsed_text.parsed_text:
          if self.line.filename[0].startswith(filename):
            break
        text = []
        same_as = self.same_as
        inside = False
        for vars,line in buffer:
          if type(line) == Begin_provider:
            if line.filename[1] == same_as:
              inside = True
            vars = []
          if inside:
            text.append( (vars,line) )
            text += map( lambda x: ([],Simple_line(line.i,x,line.filename)), call_provides(vars) )
            if command_line.do_profile and type(line) == Begin_provider:
              text.append( ( [], Declaration(line.i,"  double precision :: irp_rdtsc, irp_rdtsc1, irp_rdtsc2",line.filename) ) )
              text.append( ( [], Simple_line(line.i,"  irp_rdtsc1 = irp_rdtsc()",line.filename) ) )
          if type(line) == End_provider:
            if inside:
              break
        name = self.name
        text = parsed_text.move_to_top(text,Declaration)
        text = parsed_text.move_to_top(text,Implicit)
        text = parsed_text.move_to_top(text,Use)
        text = map(lambda x: x[1], text)
#        inside_omp = False
        for line in filter(lambda x: type(x) not in [ Begin_doc, End_doc, Doc], text):
          if type(line) == Begin_provider:
            result = []
            if command_line.directives and command_line.inline in ["all","builders"]:
              result += [ "!DEC$ ATTRIBUTES INLINE :: bld_%s"%(same_as) ]
            result += [ "subroutine bld_%s"%(name) ]
            result += build_use([name]+self.needs)
          elif type(line) == Cont_provider:
            pass
          elif type(line) == End_provider:
            if command_line.do_profile:
              result += [ "  irp_rdtsc2 = irp_rdtsc()" ,
                          "  call irp_set_timer(%d,(irp_rdtsc2-irp_rdtsc1))"%self.label ]
            result.append( "end subroutine bld_%s"%(name) )
            break
          elif type(line) == Openmp:
            # Detect OpenMP blocks
            buffer = line.text.lower().split()
            if buffer[1] == "parallel":
#              inside_omp = True
              self._has_openmp = True
#            if buffer[1] == "end" and buffer[2] == "parallel":
#              inside_omp = False
            result.append(line.text)
          else:
#           if inside_omp:
#             if type(line) in [ Provide_all, Provide, Touch, SoftTouch ]:
#               error.fail(line,str(type(line))+"is not allowed in an OpenMP block.")
            result.append(line.text)
        self._builder = result
    return self._builder
  builder = property(builder)

  ##########################################################
  def children(self):
    if '_children' not in self.__dict__:
      if not self.is_main:
        self._children = []
      from variables import variables
      if '_needs' not in self.__dict__:
        import parsed_text
      result = []
      for x in self.needs:
        result.append(x)
        try:
          result += variables[x].children
        except RuntimeError:
          pass # Exception will be checked after
      self._children = make_single(result)
      if self.name in result:
        error.fail(self.line,"Cyclic dependencies:\n%s"%(str(self._children)))
    return self._children
  children = property(children)

  ##########################################################
  def parents(self):
    if '_parents' not in self.__dict__:
      if not self.is_main:
        self._parents = []
      else:
        from variables import variables
        if '_needed_by' not in self.__dict__:
          import parsed_text
        result = []
        for x in self.needed_by:
          result.append(x)
          try:
            result += variables[x].parents
          except RuntimeError:
            pass # Exception will be checked after
        self._parents = make_single(result)
        if self.name in result:
          error.fail(self.line,"Cyclic dependencies:\n%s"%(str(self._parents)))
    return self._parents
  parents = property(parents)

######################################################################
if __name__ == '__main__':
  from preprocessed_text import preprocessed_text
  from variables import variables
 #for v in variables.keys():
 #  print v
  def print_dot(x,done):
    if x.children == []:
       return
    for i in x.needs:
       pair =  (x.name, i)
       if pair not in done:
         print "%s -> %s"%( x.name, i )
         done[pair] = None
         print_dot(variables[i],done)

  print "digraph G { "
#  print_dot(variables['e_loc'], {}) 
  print_dot(variables['psi_value'], {}) 
  print "}"



#!/usr/bin/env python2
#   IRPF90 is a Fortran90 preprocessor written in Python for programming using
#   the Implicit Reference to Parameters (IRP) method.
#   Copyright (C) 2009 Anthony SCEMAMA 
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License along
#   with this program; if not, write to the Free Software Foundation, Inc.,
#   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
#   Anthony Scemama
#   LCPQ - IRSAMC - CNRS
#   Universite Paul Sabatier
#   118, route de Narbonne      
#   31062 Toulouse Cedex 4      
#   scemama@irsamc.ups-tlse.fr


from irpf90_t import *
from util import *
from variables import variables
from modules import modules

CHECKPOINT_UNIT_NUMBER=63

FILENAME=irpdir+'irp_checkpoint.irp.F90'

def create():
  out_write      = [ "subroutine irp_checkpoint_write" ] 
  l = variables.keys()
  l.sort
  main_modules = filter(lambda x: modules[x].is_main, modules)
  for m in filter(lambda x: not modules[x].is_main, modules):
    out_write   += [ "  use %s"%(modules[m].name) ]
  out_write     += [ "  implicit none" ]
  out_write     += [ "  integer, parameter :: iunit = %d"%(CHECKPOINT_UNIT_NUMBER) ]
  out_write     += [ "  open(unit=%d,file='irp_checkpoint.dat',status='UNKNOWN',action='WRITE')"%(CHECKPOINT_UNIT_NUMBER) ]
  for v in l:
    var = variables[v]
    if var.is_main:
      out_write     += [ "  if (%s_is_built) then"%(v) ]
      for w in [v]+var.others:
        d = variables[w].dim
        if d == []:
          out_write += [ "    write(iunit,*) '%s', 0"%(w) ]
        else:
          out_write += [ "    write(iunit, *) '%s', %d"%(w, len(d)),  
                        "    write(iunit, *) %s"%(",".join(
                        [ "size(%s,%d)"%(w,i+1) for i in range(len(d)) ] ))
                       ]
        out_write   += [ "    write(iunit,*) %s"%(w) ]
      out_write     += [ "  endif" ]
  out_write         += [ "  close(%d)"%(CHECKPOINT_UNIT_NUMBER) ]
  out_write         += [ "end" ]

  out = '\n'.join(out_write)
  if not same_file(FILENAME,out):
    file = open(FILENAME,'w')
    file.writelines(out)
    file.close()

if __name__ == '__main__':
  create()




#!/usr/bin/env python2
#   IRPF90 is a Fortran90 preprocessor written in Python for programming using
#   the Implicit Reference to Parameters (IRP) method.
#   Copyright (C) 2009 Anthony SCEMAMA 
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License along
#   with this program; if not, write to the Free Software Foundation, Inc.,
#   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
#   Anthony Scemama
#   LCPQ - IRSAMC - CNRS
#   Universite Paul Sabatier
#   118, route de Narbonne      
#   31062 Toulouse Cedex 4      
#   scemama@irsamc.ups-tlse.fr


import re

re_comment = re.compile(r"^([^'!]*)('[^']*'[^']*)*!")

re_decl    = re.compile( "".join( [ r"^\ *",
 r"(integer[(::)?\* ,]+",
 r"|double *precision[(::)?\* ,]+",
 r"|logical[(::)?\* ,]+",
 r"|character[(::)?\* ,]+",
 r"|real[(::)?\* ,]+",
 r"|dimension[(::)?\* ,]+",
 r"|parameter[(::)?\* ,]+",
 r"|data */",
 r"|allocatable *(::)?",
 r"|common */",
 r"|namelist */",
 r"|save */",
 r"|complex[(::)?\* ,]+",
 r"|intrinsic *(::)?",
 r"|external *(::)?",
 r"|equivalence *(::)?",
 r"|type",
 r"|end ?type",
 r")[^=(]"
] ) )

re_test  = re.compile(r"\( *(.*)(\.[a-zA-Z]*\.|[<>]=?|[=/]=)([^=]*)\)")

re_string = re.compile(r"'.*?'")




#!/usr/bin/env python2
#   IRPF90 is a Fortran90 preprocessor written in Python for programming using
#   the Implicit Reference to Parameters (IRP) method.
#   Copyright (C) 2009 Anthony SCEMAMA 
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License along
#   with this program; if not, write to the Free Software Foundation, Inc.,
#   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
#   Anthony Scemama
#   LCPQ - IRSAMC - CNRS
#   Universite Paul Sabatier
#   118, route de Narbonne      
#   31062 Toulouse Cedex 4      
#   scemama@irsamc.ups-tlse.fr


import os,sys
import irpf90_t
from command_line import command_line
from modules import modules
irpdir = irpf90_t.irpdir
mandir = irpf90_t.mandir
irp_id = irpf90_t.irp_id

FILENAME = os.path.join(irpdir,"build.ninja")

cwd = os.getcwd()

PRINT_WIDTH=50

def dress(f,in_root=False):
    """
    Transfoms the filename f into $PWD/IRPF90_temp/f
    """
    if in_root:
      result = os.path.join(cwd,f)
    else:
      result = os.path.join(cwd,irpdir,f)
    return os.path.normpath(result)


def create_build_touches(list_of_other_o):
    """
    Create the build command for the irp_touches.o file and the irpf90.a library.
    """
    name = "irp_touches"
    short_lib  = "irpf90.a"
    short_target_o    = "%s.irp.o"%name
    short_target_F90  = "%s.irp.F90"%name
    lib               = dress(short_lib)
    target_o          = dress(short_target_o)
    target_F90        = dress(short_target_F90)

    needed_modules = [ "%s.irp.module.o"%(modules[x].filename) for x in modules ]

    list_of_modules = map(dress, needed_modules) + list_of_other_o
    list_of_modules = ' '.join(list_of_modules)

    result = '\n'.join(
        [ 
          "build {target_o}: compile_touches_{id} {target_F90} | {list_of_modules}",
          "   short_in = {short_target_F90}",
          "   short_out = {short_target_o}",
          "",
          "build {lib}: link_lib_{id} {target_o} {list_of_modules}",
          "   short_out = {short_lib}",
          "",
        ] )

    result = result.format(
        id                = irp_id            ,
        lib               = lib               ,
        list_of_modules   = list_of_modules   ,
        short_lib         = short_lib         ,
        short_target_F90  = os.path.split(target_F90)[1]  ,
        short_target_o    = os.path.split(target_o)[1]    ,
        target_F90        = target_F90        ,
        target_o          = target_o           
      )
    return result

def create_build_target(t,list_of_other_o):
    """
    Create the build command for the target module t. t is a Fmodule object.
    """
    name = t.filename

    irp_lib           = dress("irpf90.a")
    target            = dress(name,in_root=True)
    short_target_o    = "%s.irp.o"%name
    short_target_F90  = "%s.irp.F90"%name
    short_target_module_F90 = "%s.irp.module.F90"%name
    short_target_module_o = "%s.irp.module.o"%name
    target_o          = dress(short_target_o)
    target_F90        = dress(short_target_F90)
    target_module_o   = dress(short_target_module_o)
    target_module_F90 = dress(short_target_module_F90)

    needed_modules = [ "%s.irp.module.o"%(modules[x].filename) for x in modules \
        if modules[x].name in t.needed_modules ] + [ target_module_o ]

    list_of_o = [ target_o, target_module_o, irp_lib ] 
    list_of_o = ' '.join(list_of_o)

    list_of_modules = map(dress, needed_modules) + list_of_other_o
    list_of_modules = ' '.join(list_of_modules)

    list_of_includes = ' '.join(map(lambda x: dress(x,in_root=True), t.includes))

    result = '\n'.join(
        [ "build {target}: link_{id} {list_of_o}", 
          "   short_out = {short_target}",
          "",
          "build {target_o}: compile_fortran_{id} {target_F90} | {list_of_modules} {list_of_includes}",
          "   short_in  = {short_target_F90}",
          "   short_out = {short_target_o}",
          "",
          "build {target_module_o}: compile_fortran_{id} {target_module_F90}",
          "   short_in  = {short_target_module_F90}",
          "   short_out = {short_target_module_o}",
          "",
        ] )

    result = result.format(
        id                      = irp_id                  ,
        list_of_includes        = list_of_includes        ,
        list_of_modules         = list_of_modules         ,
        list_of_o               = list_of_o               ,
        short_target_F90        = os.path.split(target_F90)[1]        ,
        short_target_module_F90 = os.path.split(target_module_F90)[1] ,
        short_target_module_o   = os.path.split(target_module_o)[1]   ,
        short_target            = name                    ,
        short_target_o          = os.path.split(target_o)[1]          ,
        target_F90              = target_F90              ,
        target_module_F90       = target_module_F90       ,
        target_module_o         = target_module_o         ,
        target_o                = target_o                ,
        target                  = target             
      )
    return result


def create_build_non_target(t,list_of_other_o):
    """
    Create the build command for the non-target module t. t is a Fmodule object.
    """
    name = t.filename

    target            = dress(name,in_root=True)
    short_target_o    = "%s.irp.o"%name
    short_target_F90  = "%s.irp.F90"%name
    short_target_module_F90 = "%s.irp.module.F90"%name
    short_target_module_o = "%s.irp.module.o"%name
    target_o          = dress(short_target_o)
    target_F90        = dress(short_target_F90)
    target_module_o   = dress(short_target_module_o)
    target_module_F90 = dress(short_target_module_F90)

    needed_modules = [ "%s.irp.module.o"%(modules[x].filename) for x in modules \
        if modules[x].name in t.needed_modules ] + [ target_module_o ]

    list_of_modules = map(dress, needed_modules)
    list_of_modules = ' '.join(list_of_modules)
    list_of_externals = ' '.join([ dress(x,in_root=True) for x in list_of_other_o ])
    list_of_includes = ' '.join(map(lambda x: dress(x,in_root=True), t.includes))

    result = '\n'.join(
        [ 
          "build {target_o}: compile_fortran_{id} {target_F90} | {list_of_modules} {list_of_externals}",
          "   short_in  = {short_target_F90}",
          "   short_out = {short_target}",
          "",
          "build {target_module_o}: compile_fortran_{id} {target_module_F90} | {list_of_externals} {list_of_includes}",
          "   short_in  = {short_target_F90}",
          "   short_out = {short_target_o}",
          "",
        ] )

    result = result.format(
        id                = irp_id            ,
        list_of_externals = list_of_externals ,
        list_of_includes  = list_of_includes  ,
        list_of_modules   = list_of_modules   ,
        short_target_F90  = os.path.split(target_F90)[1]  ,
        short_target      = name              ,
        short_target_o    = os.path.split(target_o)[1]    ,
        target_F90        = target_F90        ,
        target_module_F90 = target_module_F90 ,
        target_module_o   = target_module_o   ,
        target_o          = target_o           
      )
    return result


def create_build_remaining(f):
    """
    Create the build command for the remaining file f. f is a file name (str).
    """
    t, extension = f.rsplit('.',1)
    t1 = dress(t,in_root=True)
    t2 = dress(t,in_root=False)
    target_i  = f
    target_o  = "%s.o"%t
    if not target_o.startswith(os.path.join(cwd,irpdir)):
      target_o = target_o.replace(cwd,os.path.join(cwd,irpdir))

    if extension.lower() in [ 'f', 'f90' ]:
      result = [ "build {target_o}: compile_fortran_{id} {target_i}" ]
    elif extension.lower() in [ 'c' ]:
      result = [ "build {target_o}: compile_c_{id} {target_i}" ]
    elif extension.lower() in [ 'cxx', 'cpp' ]:
      result = [ "build {target_o}: compile_cxx_{id} {target_i}" ]

    result += [ "   short_in  = {short_target_i}", 
                "   short_out = {short_target_o}", "" ]
    result = '\n'.join(result).format(
        target_o  = target_o,
        target_i  = target_i,
        short_target_o  = os.path.split(target_o)[1],
        short_target_i  = os.path.split(target_i)[1],
        id        = irp_id
      )
    return result



def create_irpf90_make(targets):
    targets = ' '.join(targets)
    result = """NINJA += -C {0}

TARGETS={1}

.PHONY: all clean veryclean

all:
	$(NINJA)

$(TARGETS): 
	$(NINJA) $(PWD)/$@

clean:
	$(NINJA) -t clean

veryclean: clean
	rm -rf IRPF90_temp/ IRPF90_man/ irpf90.make irpf90_entities dist tags

""".format(irpdir, targets)

    import makefile
    with open(makefile.IRPF90_MAKE,'w') as file:
        file.write(result)


def run():

    output = [ "builddir = %s"%os.path.join(cwd,irpdir) ]

    # Environment variables

    try: FC = os.environ["FC"]
    except KeyError: FC="gfortran -ffree-line-length-none"

    try: AR = os.environ["AR"]
    except KeyError: AR="ar"

    try: CC = os.environ["CC"]
    except KeyError: CC="gcc"

    try: CXX = os.environ["CXX"]
    except KeyError: CXX="g++"

    includes = [ "-I %s"%(i) for i in command_line.include_dir ]

    FC  += " "+' '.join(includes)
    CC  += " "+' '.join(includes)
    CXX += " "+' '.join(includes)
   
    try:  SRC = os.environ["SRC"].split()
    except KeyError: SRC=[]

    try:  OBJ = os.environ["OBJ"].split()
    except KeyError: OBJ=[]

    try:  LIB = os.environ["LIB"].split()
    except KeyError: LIB=[]

    try: FCFLAGS = os.environ["FCFLAGS"].split()
    except KeyError: FCFLAGS = [ "-O2" ]

    try: CFLAGS = os.environ["CFLAGS"].split()
    except KeyError: CFLAGS = [ "-O2" ]

    try: CXXFLAGS = os.environ["CXXFLAGS"].split()
    except KeyError: CXXFLAGS = [ "-O2" ]

    FCFLAGS = ' '.join(FCFLAGS)
    CFLAGS  = ' '.join(CFLAGS)
    CXXFLAGS  = ' '.join(CXXFLAGS)
    LIB = ' '.join(LIB)

    # Rules

    t = [ "rule compile_fortran_{id}", 
          "  command = {FC} {FCFLAGS} -c $in -o $out", 
          "  description = F   : $short_in -> $short_out", 
          "",
          "rule compile_touches_{id}", 
          "  command = {FC} -c $in -o $out",
          "  description = F   : $short_in -> $short_out", 
          "",
          "",
          "rule compile_c_{id}" ,
          "  command = {CC} {CFLAGS} -c $in -o $out", 
          "  description = C   : $short_in -> $short_out", 
          "",
          "rule compile_cxx_{id}", 
          "  command = {CXX} {CXXFLAGS} -c $in -o $out", 
          "  description = C++ :  $short_in -> $short_out", 
          "",
          "rule link_lib_{id}", 
          "  command = {AR} crf $out $in" ,
          "  description = Link: $short_out", 
          "",
          "rule link_{id}", 
          "  command = {FC} $in {LIB} -o $out" ,
          "  description = Link: $short_out"] 
   
    output += [ '\n'.join(t).format(id=irp_id, FC=FC, FCFLAGS=FCFLAGS, LIB=LIB, CXX=CXX, CXXFLAGS=CXXFLAGS, CC=CC, CFLAGS=CFLAGS, AR=AR)  ]


    # All modules : list of Fmodule objects

    l_mod = list( modules.values() )


    # Modules that are not targets

    l_non_targets = [ x for x in l_mod if not x.is_main ]




    # Common object/source files

    l_common_o = [ "irp_touches.irp.o" ] + \
      [ "{0}.irp.o".format(x.filename) for x in l_non_targets ] + \
      [ "{0}.irp.module.o".format(x.filename) for x in l_non_targets ]


    l_common_s = []

    if command_line.do_debug:
        l_common_o += [ "irp_stack.irp.o" ]
        l_common_s += [ "irp_stack.irp.F90" ]

    if command_line.do_openmp:
        l_common_o += [ "irp_locks.irp.o" ]
        l_common_s += [ "irp_locks.irp.F90" ]

    if command_line.do_profile:
        l_common_o += [ "irp_profile.irp.o", "irp_rdtsc.o" ]
        l_common_s += [ "irp_profile.irp.F90", "irp_rdtsc.c" ]

    l_common_o = map(dress,l_common_o) + map(lambda x: dress(x,in_root=True), OBJ)
    l_common_s = map(dress,l_common_s) + map(lambda x: dress(x,in_root=True), SRC)


    # IRP_touches
    output.append(create_build_touches(l_common_o[1:]))

    # All targets : list of Fmodule objects

    l_targets = [ x for x in l_mod if x.is_main ]


    # Create lines

    for i in l_non_targets:
        output.append (create_build_non_target(i,OBJ))

    for i in l_targets:
        output.append(create_build_target(i, l_common_o))

    # Remaining files
    for i in l_common_s:
        output.append(create_build_remaining(i))

    
    with open(FILENAME,'w') as f:
        f.write('\n\n'.join(output))
        f.write('\n')

    create_irpf90_make([ x.filename for x in l_targets ] + [ os.path.join(irpdir,'irpf90.a') ] )

    return





#!/usr/bin/env python2
#   IRPF90 is a Fortran90 preprocessor written in Python for programming using
#   the Implicit Reference to Parameters (IRP) method.
#   Copyright (C) 2009 Anthony SCEMAMA 
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License along
#   with this program; if not, write to the Free Software Foundation, Inc.,
#   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
#   Anthony Scemama
#   LCPQ - IRSAMC - CNRS
#   Universite Paul Sabatier
#   118, route de Narbonne      
#   31062 Toulouse Cedex 4      
#   scemama@irsamc.ups-tlse.fr


import os
import sys

wd = os.path.abspath(os.path.dirname(__file__))
from irpf90_t import mandir
filename = sys.argv[1].lower()+".l"
if filename not in os.listdir(mandir):
   print "%s does not exist"%(sys.argv[1])
   sys.exit(-1)

os.system("man ./"+mandir+sys.argv[1].lower()+".l")




#!/usr/bin/env python2
#   IRPF90 is a Fortran90 preprocessor written in Python for programming using
#   the Implicit Reference to Parameters (IRP) method.
#   Copyright (C) 2009 Anthony SCEMAMA 
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License along
#   with this program; if not, write to the Free Software Foundation, Inc.,
#   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
#   Anthony Scemama
#   LCPQ - IRSAMC - CNRS
#   Universite Paul Sabatier
#   118, route de Narbonne      
#   31062 Toulouse Cedex 4      
#   scemama@irsamc.ups-tlse.fr


from command_line import command_line

from irpf90_t import *
from util import *
from variables import variables
FILENAME=irpdir+'irp_locks.irp.F90'

def create():
  out = []
  l = variables.keys()
  l.sort
  for v in l:
    var = variables[v]
    out += var.locker

  out += [
"subroutine irp_init_locks_%s()"%(irp_id),
" implicit none" ]
  for v in l:
    out += [ "  call irp_lock_%s(.True.)"%v ]
    out += [ "  call irp_lock_%s(.False.)"%v ]
  out += [ "end subroutine" ]
  out = map(lambda x: "%s\n"%(x),out)
  if not same_file(FILENAME,out):
    file = open(FILENAME,'w')
    file.writelines(out)
    file.close()

if __name__ == '__main__':
  create()
  



#!/usr/bin/env python2
#   IRPF90 is a Fortran90 preprocessor written in Python for programming using
#   the Implicit Reference to Parameters (IRP) method.
#   Copyright (C) 2009 Anthony SCEMAMA 
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License along
#   with this program; if not, write to the Free Software Foundation, Inc.,
#   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
#   Anthony Scemama
#   LCPQ - IRSAMC - CNRS
#   Universite Paul Sabatier
#   118, route de Narbonne      
#   31062 Toulouse Cedex 4      
#   scemama@irsamc.ups-tlse.fr

import os
NTHREADS=1 #int(os.getenv('OMP_NUM_THREADS',1))

def strip(x):
  return x.strip()

def lower(x):
  return x.lower()

def same_file(filename,txt):
  assert isinstance(filename,str)
  if (type(txt) == list):
    buffer = ''.join(txt)
  else:
    buffer = txt

  try:
    file = open(filename,"r")
  except IOError:
    return False
  stream = file.read()
  file.close()

  if len(stream) != len(buffer):
    return False
  if stream != buffer:
    return False
  return True

def build_dim(dim):
  if len(dim) == 0:
    return ""
  else:
    return "(%s)"%( ",".join(dim) )

def build_dim_colons(v):
  d = v.dim
  if d == []:
    return ""
  else:
    x = map(lambda x: ":", d)
    return "(%s)"%(','.join(x))


import error
def find_subname(line):
  buffer = line.lower
  if not buffer.endswith(')'):
    buffer += "()"
  buffer = buffer.split('(')
  buffer = buffer[0].split()
  if len(buffer) < 2:
    error.fail(line,"Syntax Error")
  return buffer[-1]

def make_single(l):
  d = {}
  for x in l:
   d[x] = True
  return d.keys()

def flatten(l):
  if type(l) == list:
    result = []
    for i in range(len(l)):
      elem = l[i]
      result += flatten(elem)
    return result
  else:
    return [l]

def dimsize(x):
  assert isinstance(x,str)
  buffer = x.split(':')
  if len(buffer) == 1:
    return x
  else:
    assert len(buffer) == 2
    size = ""
    b0, b1 = buffer
    if b0.replace('-','').isdigit() and b1.replace('-','').isdigit():
      size = str( int(b1) - int(b0) + 1 )
    else:
      if b0.replace('-','').isdigit():
        size = "(%s) - (%d)"%(b1,int(b0)-1)
      elif b1.replace('-','').isdigit():
        size = "(%d) - (%s)"%(int(b1)+1,b0)
      else:
        size = "(%s) - (%s) + 1"%(b1,b0)
    return size

def put_info(text,filename):
  assert type(text) == list
  if len(text) > 0:
    assert type(text[0]) == tuple
    from irpf90_t import Line
    assert type(text[0][0]) == list
    assert isinstance(text[0][1], Line)
    lenmax = 80 - len(filename)
    format = "%"+str(lenmax)+"s ! %s:%4s"
    for vars,line in text:
      line.text = format%(line.text.ljust(lenmax),line.filename,str(line.i))
  return text

import cPickle as pickle
import os, sys
def parallel_loop(f,source):
  pidlist = range(NTHREADS)

  src = [ [] for i in xrange(NTHREADS) ]
  index = 0
  try:
    source = map( lambda x: (len(x[1]),(x[0], x[1])), source )
    source.sort()
    source = map( lambda x: x[1], source )
  except:
    pass
  for i in source:
    index += 1
    if index == NTHREADS:
      index = 0
    src[index].append(i)

  thread_id = 0
  fork = 1
  r = range(0,NTHREADS)
  for thread_id in xrange(1,NTHREADS):
    r[thread_id], w = os.pipe()
    fork = os.fork()
    if fork == 0:
      os.close(r[thread_id])
      w = os.fdopen(w,'w')
      break
    else:
      os.close(w)
      r[thread_id] = os.fdopen(r[thread_id],'r')
      pidlist[thread_id] = fork
      thread_id = 0

  result = []
  for filename, text in src[thread_id]:
    result.append( (filename, f(filename,text)) )
  result.sort()

  if fork == 0:
    pickle.dump(result,w,-1)
    w.close()
    os._exit(0)
  
  for i in xrange(1,NTHREADS):
    result += pickle.load(r[i])
    r[i].close()
    os.waitpid(pidlist[i],0)[1] 

  return result



if __name__ == '__main__':
  print "10",dimsize("10") #-> "10"
  print "0:10",dimsize("0:10") # -> "11"
  print "0:x",dimsize("0:x")  # -> "x+1"
  print "-3:x",dimsize("-3:x")  # -> "x+1"
  print "x:y",dimsize("x:y")  # -> "y-x+1"
  print "x:5",dimsize("x:5")  # -> "y-x+1"

